from __future__ import annotations

import copy
import re
import statistics
import functools
import yaml
import tempfile
import contextlib
import pathlib
import math
import os
import datetime as datetime_module
import lzma
import random
import json
import shutil

import pandas as pd
import sys
import subprocess
import itertools
import operator
import collections.abc
import contextlib
import pathlib
import pickle
import more_itertools
from typing import (Any, Mapping, Optional, BinaryIO, Iterable, TypeVar, TextIO, Iterator, Generator,
                    Sequence)
import io

import numpy as np
import ray.rllib
import svgwrite

from .constants import CHICKEN_COOP_HOME
from chicken_coop.county import csv_tools
from .typing import Agent, Action, RealNumber
from chicken_coop import county
from . import filtros





def sane_kwargs(function):
    @functools.wraps(function)
    def inner(config):
        return function(**config)
    return inner


def jsonl_to_dataframe(jsonl_path: str | pathlib.Path, *, dot_notation: bool = False) -> pd.DataFrame:
    jsonl_path = pathlib.Path(jsonl_path)
    with jsonl_path.open() as file:
        entries = list(map(json.loads, file))
    if dot_notation:
        for entry in entries:
            unprocessed_keys = set(entry)
            while unprocessed_keys:
                key = unprocessed_keys.pop()
                if isinstance(entry[key], dict):
                    for sub_key, sub_value in entry[key].items():
                        unprocessed_keys.add(dotted_key := f'{key}.{sub_key}')
                        entry[dotted_key] = sub_value
                    del entry[key]

    return pd.DataFrame(entries)


def compute_actions_for_all_agents(algorithm: ray.rllib.algorithms.Algorithm,
                                   env: county.BaseEnv) -> dict[Agent, Action]:
    return {agent: algorithm.compute_single_action(
        env.observation_by_agent[agent],
        policy_id=algorithm.config['policy_mapping_fn'](agent)
    )
            for agent in env.get_agent_ids()}



def compickle(thing: Any, file_or_path: pathlib.Path | BinaryIO, /) -> None:
    with contextlib.ExitStack() as exit_stack:
        if isinstance(file_or_path, pathlib.Path):
            file_ = exit_stack.enter_context(file_or_path.open('wb'))
        else:
            file_ = file_or_path
        file_.write(lzma.compress(pickle.dumps(thing)))


def compickle_to_bytes(thing: Any)-> bytes:
    bytes_io = io.BytesIO()
    compickle(thing, bytes_io)
    return bytes_io.getvalue()

def uncompickle(file_or_path: pathlib.Path | BinaryIO, /) -> Any:
    with contextlib.ExitStack() as exit_stack:
        if isinstance(file_or_path, pathlib.Path):
            file_ = exit_stack.enter_context(file_or_path.open('rb'))
        else:
            file_ = file_or_path
        return pickle.loads(lzma.decompress(file_.read()))

def uncompickle_from_bytes(compickled_bytes: bytes) -> Any:
    return uncompickle(io.BytesIO(compickled_bytes))


class TroubleshootingDict(collections.abc.MutableMapping):
    def __init__(self, *args, **kwargs):
        self._map = dict(*args, **kwargs)

    def __getitem__(self, key):
        return self._map[key]

    def __delitem__(self, key):
        del self._map[key]

    def __setitem__(self, key, value):
        if isinstance(value, list):
            if all(isinstance(item, list) for item in value):
                value = TroubleshootingList(map(TroubleshootingList, value))
            else:
                value = TroubleshootingList(value)
        self._map[key] = value

    def __len__(self):
        return len(self._map)

    def __iter__(self):
        return iter(self._map)

    def __repr__(self):
        return f'{type(self).__name__}({self._map})'


class TroubleshootingList(collections.abc.MutableSequence):
    def __init__(self, *args, **kwargs):
        self._list = list(*args, **kwargs)

    def __getitem__(self, key):
        return self._list[key]

    def __delitem__(self, key):
        del self._list[key]

    def __setitem__(self, key, value):
        self._list[key] = value

    def __len__(self):
        return len(self._list)

    def __iter__(self):
        return iter(self._list)

    def insert(self, index, thing):
        self._list.insert(index, thing)

    def __repr__(self):
        return f'{type(self).__name__}({self._list})'



def all_equal(values: Iterable[Any]) -> bool:
    return all(itertools.starmap(operator.eq, more_itertools.windowed(values, 2)))


def cute_div(x: RealNumber, y: RealNumber, *, default: RealNumber = 0) -> RealNumber:
    try:
        return x / y
    except ZeroDivisionError:
        return default

# def cool_windowed(iterable: Iterable[Any], n: int = 2) -> Iterable[tuple[Any, Any]]:
    # sentinel = object()
    # for left, right in more_itertools.windowed(iterable, n, fillvalue=sentinel):
        # if sentinel in (left, right):
            # return
        # yield (left, right)

def get_chicken_coop_repo_commit() -> Optional[str]:
    try:
        result = subprocess.run(
            ('/usr/bin/git', 'rev-parse', 'HEAD'),
            check=True,
            cwd=pathlib.Path(__file__).parent,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            encoding='utf-8',
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return None

def get_chicken_coop_repo_is_clean() -> Optional[bool]:
    result = subprocess.run(
        (cmd := ('/usr/bin/git', 'diff-index', '--quiet', 'HEAD')),
        cwd=pathlib.Path(__file__).parent,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding='utf-8',
    )
    if result.returncode == 0:
        return True
    elif result.returncode == 1:
        return False
    else:
        return None

def clamp(number: RealNumber, low: RealNumber, high: RealNumber) -> RealNumber:
    if number < low:
        return low
    elif number > high:
        return high
    else:
        return number


class TeeStream:
    def __init__(self, original_stream: TextIO, path: pathlib.Path)-> None:
        self.original_stream = original_stream
        self.path = path

    def write(self, message: str) -> None:
        self.original_stream.write(message)
        with self.path.open('a') as file:
            file.write(message)

    def flush(self) -> None:
        self.original_stream.flush()

    def fileno(self) -> int:
        return self.original_stream.fileno()

    def close(self) -> None:
        pass


@contextlib.contextmanager
def tee_stdout(path: pathlib.Path) -> None:
    original_stdout = sys.stdout
    sys.stdout = TeeStream(original_stdout, path)
    try:
        yield
    finally:
        sys.stdout = original_stdout


@contextlib.contextmanager
def tee_stderr(path: pathlib.Path, *, ensure_filtros: bool = True) -> None:
    original_stderr = sys.stderr
    sys.stderr = TeeStream(original_stderr, path)
    try:
        if ensure_filtros:
            filtros.activate()
        yield
    finally:
        sys.stderr = original_stderr


class BaseRolloutReporter(collections.abc.Sequence):
    def __init__(self) -> None:
        self.rows = []

    def __iter__(self) -> Iterator[dict[str, Any]]:
        yield from self.rows

    def __getitem__(self, i: int) -> dict[str, Any]:
        return self.rows[i]

    def __len__(self) -> int:
        return len(self.rows)

    def __reversed__(self) -> Iterator[dict[str, Any]]:
        yield from reversed(self.rows)

    def as_dataframe(self) -> pd.DataFrame:
        pd.DataFrame.from_dict(self.rows)

    def report(self, row_or_rows: Mapping[str, Any] | Iterable[Mapping[str, Any]], /) -> None:
        for row in self._parse_row_or_rows(row_or_rows):
            row = dict(row)
            self.rows.append(row)

    @staticmethod
    def _parse_row_or_rows(row_or_rows: Mapping[str, Any] | Iterable[Mapping[str, Any]]
                           ) -> Iterable[Mapping[str, Any]]:
        if isinstance(row_or_rows, collections.abc.Mapping):
            return (row_or_rows,)
        else:
            return row_or_rows

class NumpyJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, np.floating):
            if np.isnan(obj):
                return None  # Serialized as JSON null.
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super().default(obj)

class TuneRolloutReporter(BaseRolloutReporter):
    def report(self, row_or_rows: Mapping[str, Any] | Iterable[Mapping[str, Any]], /) -> None:
        import ray.tune
        for row in self._parse_row_or_rows(row_or_rows):
            BaseRolloutReporter.report(self, row)
            ray.tune.report(**row)

class BaseFileRolloutReporter(BaseRolloutReporter):
    def __init__(self, path: pathlib.Path) -> None:
        BaseRolloutReporter.__init__(self)
        self.path = path


class CsvRolloutReporter(BaseFileRolloutReporter):
    def report(self, row_or_rows: Mapping[str, Any] | Iterable[Mapping[str, Any]], /) -> None:
        for row in self._parse_row_or_rows(row_or_rows):
            BaseFileRolloutReporter.report(self, row)
            with csv_tools.CSVWriter(self.path, row.keys(),
                                     overwrite=(len(self.rows) == 1)) as csv_writer:
                csv_writer.write_row(row)

class JsonlRolloutReporter(BaseFileRolloutReporter):
    def report(self, row_or_rows: Mapping[str, Any] | Iterable[Mapping[str, Any]], /) -> None:
        for row in self._parse_row_or_rows(row_or_rows):
            BaseFileRolloutReporter.report(self, row)
            make_tree = lambda: collections.defaultdict(make_tree)
            tree = make_tree()
            for key, value in row.items():
                *parents, short_name = key.split('.')
                subtree = tree
                for parent in parents:
                    subtree = subtree[parent]
                subtree[short_name] = value

            with self.path.open('a') as file:
                json.dump(tree, file, cls=NumpyJsonEncoder)
                file.write('\n')



def shuffled(iterable: Iterable[Any], *, seed: Optional[int] = None) -> tuple[Any, ...]:
    list_ = list(iterable)
    random.Random(seed).shuffle(list_)
    return tuple(list_)

def get_mean_dataframe_from_experiment_analysis(
    experiment_analysis: ray.tune.ExperimentAnalysis,
    fields_that_need_last: tuple[str] = ()) -> pd.DataFrame:

    rows = {
        path: df.mean(numeric_only=True).to_dict()
        for path, df in experiment_analysis.trial_dataframes.items()
    }

    for field_that_needs_last in fields_that_need_last:
        for path, df in experiment_analysis.trial_dataframes.items():
            rows[path][field_that_needs_last] = df[field_that_needs_last].values[-1]

    all_configs = experiment_analysis.get_all_configs(prefix=True)
    for path, config in all_configs.items():
        if path in rows:
            rows[path].update(config)
            rows[path].update(logdir=path)
    return pd.DataFrame(list(rows.values()))

def int_sqrt(x: RealNumber) -> int:
    sqrt_float = math.sqrt(x)
    sqrt_int = int(sqrt_float)
    if not sqrt_int == sqrt_float:
        raise ValueError(f"Got number {x} when expecting a number that's a square of an "
                         f"integer. Alas, the sqrt of {x} is a non-int: {sqrt_float}")
    return sqrt_int


def make_one_hot(i: int, n: int) -> np.ndarray:
    array = np.zeros((n,), dtype=np.int8)
    array[i] = 1
    return array

@contextlib.contextmanager
def create_temp_folder(*, prefix=tempfile.template, suffix='',
                       parent_folder=None, chmod=None):
    '''
    Context manager that creates a temporary folder and deletes it after usage.

    After the suite finishes, the temporary folder and all its files and
    subfolders will be deleted.

    Example:

        with create_temp_folder() as temp_folder:

            # We have a temporary folder!
            assert temp_folder.is_dir()

            # We can create files in it:
            (temp_folder / 'my_file').open('w')

        # The suite is finished, now it's all cleaned:
        assert not temp_folder.exists()

    Use the `prefix` and `suffix` string arguments to dictate a prefix and/or a
    suffix to the temporary folder's name in the filesystem.

    If you'd like to set the permissions of the temporary folder, pass them to
    the optional `chmod` argument, like this:

        create_temp_folder(chmod=0o550)

    '''
    temp_folder = pathlib.Path(tempfile.mkdtemp(prefix=prefix, suffix=suffix,
                                                dir=parent_folder))
    try:
        if chmod is not None:
            temp_folder.chmod(chmod)
        yield temp_folder
    finally:
        shutil.rmtree(str(temp_folder))

def zero_based_accumulate(iterable: Iterable[RealNumber],
                          drop_last: bool = False) -> Iterable[RealNumber]:
    old_item = 0
    for new_item in itertools.accumulate(iterable):
        yield old_item
        old_item = new_item
    if not drop_last:
        yield new_item


def duplicate_shape_with_attribs(shape: svgwrite.base.BaseElement, **attribs
                                 ) -> svgwrite.base.BaseElement:
    new_shape = copy.deepcopy(shape)
    new_shape.attribs |= attribs
    return new_shape


def duplicate_with_offset(shape: svgwrite.base.BaseElement | Shapino, dx: float, dy: float
               ) -> svgwrite.base.BaseElement | Shapino:
    if isinstance(shape, Shapino):
        return shape.add_offset(dx, dy)

    attribs = shape.attribs.copy()
    match shape:
        case svgwrite.shapes.Rect():
            return duplicate_shape_with_attribs(
                shape,
                x=float(attribs.get('x', 0)) + dx,
                y=float(attribs.get('y', 0)) + dy,
            )
        case svgwrite.shapes.Circle():
            return duplicate_shape_with_attribs(
                shape,
                cx=float(attribs.get('cx', 0)) + dx,
                cy=float(attribs.get('cy', 0)) + dy,
            )
        case svgwrite.shapes.Line():
            return duplicate_shape_with_attribs(
                shape,
                x1=float(attribs.get('x1', 0)) + dx,
                y1=float(attribs.get('y1', 0)) + dy,
                x2=float(attribs.get('x2', 0)) + dx,
                y2=float(attribs.get('y2', 0)) + dy,
            )
        case svgwrite.shapes.Ellipse():
            return duplicate_shape_with_attribs(
                shape,
                cx=float(attribs.get('cx', 0)) + dx,
                cy=float(attribs.get('cy', 0)) + dy,
            )
        case svgwrite.shapes.Polyline():
            return duplicate_shape_with_attribs(
                shape,
                points=[(float(p[0]) + dx, float(p[1]) + dy,) for p in attribs.get('points', [])]
            )
        case svgwrite.shapes.Polygon():
            return duplicate_shape_with_attribs(
                shape,
                points=[(float(p[0])+dx, float(p[1])+dy) for p in attribs.get('points', [])]
            )
        case svgwrite.text.Text():
            return duplicate_shape_with_attribs(
                shape,
                x=float(attribs.get('x', 0)) + dx,
                y=float(attribs.get('y', 0)) + dy,
            )
        case svgwrite.image.Image():
            return duplicate_shape_with_attribs(
                shape,
                x=float(attribs.get('x', 0)) + dx,
                y=float(attribs.get('y', 0)) + dy,
            )
        case svgwrite.container.Use():
            return duplicate_shape_with_attribs(
                shape,
                x=float(attribs.get('x', 0)) + dx,
                y=float(attribs.get('y', 0)) + dy,
            )
        case _:
            return duplicate_shape_with_attribs(shape)

class Shapino:

    padding = 0

    def __init__(self, shapes_or_shapinos: Iterable[svgwrite.base.BaseElement | Shapino] = (),
                 x: RealNumber = 0, y: RealNumber = 0) -> None:
        self.shapes = []
        self.add(shapes_or_shapinos, x, y)

    def add(self, shapes_or_shapinos: Iterable[svgwrite.base.BaseElement | Shapino] = (),
            x: RealNumber = 0, y: RealNumber = 0) -> None:
        items = collections.deque(shapes_or_shapinos)
        while items:
            item = items.popleft()
            if isinstance(item, svgwrite.base.BaseElement):
                self.shapes.append(duplicate_with_offset(item, x, y))
            else:
                items.extend(item)

    @property
    def bounding_box(self) -> Tuple[float, float, float, float]:
        min_x = min_y = float('inf')
        max_x = max_y = float('-inf')

        for shape in self.shapes:
            match shape:
                case svgwrite.shapes.Rect():
                    x, y, w, h = map(float, (shape.attribs.get(attr, 0)
                                             for attr in ('x', 'y', 'width', 'height')))
                    min_x, min_y = min(min_x, x), min(min_y, y)
                    max_x, max_y = max(max_x, x + w), max(max_y, y + h)
                case svgwrite.shapes.Circle():
                    cx, cy, r = map(float, (shape.attribs.get(attr, 0)
                                            for attr in ('cx', 'cy', 'r')))
                    r += shape.attribs.get('stroke_width', 1) / 2
                    min_x, min_y = min(min_x, cx - r), min(min_y, cy - r)
                    max_x, max_y = max(max_x, cx + r), max(max_y, cy + r)
                case svgwrite.shapes.Line():
                    x1, y1, x2, y2 = map(float, (shape.attribs.get(attr, 0)
                                                 for attr in ('x1', 'y1', 'x2', 'y2')))
                    min_x, min_y = min(min_x, min(x1, x2)), min(min_y, min(y1, y2))
                    max_x, max_y = max(max_x, max(x1, x2)), max(max_y, max(y1, y2))
                case svgwrite.shapes.Ellipse():
                    cx, cy, rx, ry = map(float, (shape.attribs.get(attr, 0)
                                                 for attr in ('cx', 'cy', 'rx', 'ry')))
                    min_x, min_y = min(min_x, cx - rx), min(min_y, cy - ry)
                    max_x, max_y = max(max_x, cx + rx), max(max_y, cy + ry)
                case svgwrite.shapes.Polyline():
                    points = shape.attribs.get('points', [])
                    xs, ys = zip(*((float(p[0]), float(p[1])) for p in points))
                    min_x, min_y = min(min_x, min(xs)), min(min_y, min(ys))
                    max_x, max_y = max(max_x, max(xs)), max(max_y, max(ys))
                case svgwrite.shapes.Polygon():
                    points = shape.attribs.get('points', [])
                    xs, ys = zip(*((float(p[0]), float(p[1])) for p in points))
                    min_x, min_y = min(min_x, min(xs)), min(min_y, min(ys))
                    max_x, max_y = max(max_x, max(xs)), max(max_y, max(ys))
                case svgwrite.path.Path():
                    pass
                case svgwrite.text.Text():
                    x, y = float(shape.attribs.get('x', 0)), float(shape.attribs.get('y', 0))
                    min_x, min_y = min(min_x, x), min(min_y, y)
                    max_x, max_y = max(max_x, x), max(max_y, y)
                case svgwrite.image.Image():
                    x, y, w, h = map(float, (shape.attribs.get(attr, 0)
                                             for attr in ('x', 'y', 'width', 'height')))
                    min_x, min_y = min(min_x, x), min(min_y, y)
                    max_x, max_y = max(max_x, x + w), max(max_y, y + h)
                case svgwrite.container.Use():
                    x, y = float(shape.attribs.get('x', 0)), float(shape.attribs.get('y', 0))
                    min_x, min_y = min(min_x, x), min(min_y, y)
                    max_x, max_y = max(max_x, x), max(max_y, y)
                case _:
                    pass

        return (min_x - self.padding,
                min_y - self.padding,
                max_x - min_x + 2 * self.padding,
                max_y - min_y + 2 * self.padding)

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {len(self.shapes)} shapes, {self.bounding_box}>'

    def __iter__(self):
        return iter(self.shapes)

    @property
    def svg_text(self, size=('100%', '100%'), **extra) -> str:
        with create_temp_folder(prefix='chicken_coop_') as temp_folder:
            drawing_path = temp_folder / 'svg.svg'
            drawing = svgwrite.Drawing(drawing_path, size=size, **extra)
            drawing.viewbox(*self.bounding_box)
            for shape in self:
                drawing.add(shape)
            drawing.save()
            return drawing_path.read_text()

    def duplicate_with_offset(self, x: RealNumber, y: RealNumber) -> Shapino:
        new_shapino: Shapino = copy.deepcopy(self)
        new_shapino.shapes[:] = (duplicate_with_offset(shape, x, y) for shape in self)
        return new_shapino

    def duplicate_with_anchor_on_middle_left(self) -> Shapino:
        x, y, w, h = self.bounding_box
        new_shapino = self.duplicate_with_offset(-x, -y - (h / 2))
        new_x, new_y, new_w, new_h = new_shapino.bounding_box
        assert abs(new_x) < 0.01
        assert abs((new_y / (new_h / 2)) + 1) < 0.01
        return new_shapino

    def duplicate_with_anchor_on_top_center(self) -> Shapino:
        x, y, w, h = self.bounding_box
        new_shapino = self.duplicate_with_offset(-x - (w / 2), -y)
        new_x, new_y, new_w, new_h = new_shapino.bounding_box
        assert abs((new_x / (new_w / 2)) + 1) < 0.01
        assert abs(new_y) < 0.01
        return new_shapino



def get_last_line_from_file(path: pathlib.Path) -> bytes:
    with path.open('rb') as f:
        f.seek(-2, 2)  # Jump to the second last byte.

        # If the file is too short (possibly just one line or empty)
        if f.tell() == -1:
            f.seek(0)  # Go to the start
        else:
            # Search for the beginning of the last line
            while f.read(1) != b'\n':
                f.seek(-2, 1)
        return f.readline()


def get_last_entry_from_jsonl(jsonl_path: pathlib.Path) -> Any:
    return json.loads(get_last_line_from_file(jsonl_path))

def to_tuple(single_or_sequence, item_type=None, item_test=None):
    '''
    Convert an item or a sequence of items into a tuple of items.

    This is typically used in functions that request a sequence of items but
    are considerate enough to accept a single item and wrap it in a tuple
    `(item,)` themselves.

    This function figures out whether the user entered a sequence of items, in
    which case it will only be converted to a tuple and returned; or the user
    entered a single item, in which case a tuple `(item,)` will be returned.

    To aid this function in parsing, you may optionally specify `item_type`
    which is the type of the items, or alternatively `item_test` which is a
    callable that takes an object and returns whether it's a valid item. These
    are necessary only when your items might be sequences themselves.

    You may optionally put multiple types in `item_type`, and each object would
    be required to match to at least one of them.
    '''
    if (item_type is not None) and (item_test is not None):
        raise Exception('You may specify either `item_type` or '
                        '`item_test` but not both.')
    if item_test is not None:
        actual_item_test = item_test
    elif item_type is not None:
        actual_item_test = \
            lambda candidate: isinstance(candidate, item_type)
    else:
        actual_item_test = None

    if actual_item_test is None:
        if isinstance(single_or_sequence, collections.abc.Sequence):
            return tuple(single_or_sequence)
        elif single_or_sequence is None:
            return tuple()
        else:
            return (single_or_sequence,)
    else: # actual_item_test is not None
        if actual_item_test(single_or_sequence):
            return (single_or_sequence,)
        elif single_or_sequence is None:
            return ()
        else:
            return tuple(single_or_sequence)


def int_to_base(n: int, base: int) -> tuple:
    if n == 0:
        return (0,)
    digits = []
    while n:
        digits.append(n % base)
        n //= base
    return tuple(digits)



class ListifyingDumper(yaml.Dumper):
    def _represent_tuple(self, data):
        return self.represent_list(data)

ListifyingDumper.add_representer(tuple, ListifyingDumper._represent_tuple)


def _get_wacky_choices(m: int, *, n: int, k: int) -> tuple[tuple, ...]:
    if n == 0:
        return ((),) * k
    result = []
    counter = collections.Counter({i: 0 for i in range(m)})
    for _ in range(k):
        least_chosen = counter.most_common()[-1][0]
        fluffs = tuple(
            sorted([least_chosen] + random.sample(tuple(range(least_chosen)) +
                                                  tuple(range(least_chosen + 1, m)), n - 1))
        )
        for fluff in fluffs:
            counter[fluff] += 1
        result.append(fluffs)

    return tuple(result)


def get_wacky_choices(iterable: Iterable, /, *, n: int, k: int) -> tuple[tuple, ...]:
    sequence = iterable if isinstance(iterable, Sequence) else tuple(iterable)
    result = _get_wacky_choices(len(sequence), n=n, k=k)
    return tuple(tuple(map(sequence.__getitem__, fluff)) for fluff in result)


def get_int_from_mixed_string(s: str) -> int:
    return int(re.fullmatch('^[^0-9]*([0-9]+)[^0-9]*$', s).group(1))


def mean_default(numbers: Iterable[float | int], *, default: Optional[float | int] = None
                 ) -> float | int:
    numbers = tuple(numbers)
    if not numbers and default is not None:
        return default
    return statistics.mean(numbers)




from __future__ import annotations

HTML_TEMPLATE = '''
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>SVG Centered</title>
    <style>
        body, html {{
            height: 100%;
            margin: 0;
            background-color: white;
            display: flex;
            justify-content: center;
            align-items: center;
        }}
        svg {{
            max-height: 100%;
            max-width: 100%;
            width: 100%;
            height: 100%;
        }}
    </style>
</head>
<body>
    <svg viewBox="0 0 100 100" preserveAspectRatio="xMidYMid meet">
        {}
    </svg>
</body>
</html>
'''



from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Iterator
import abc
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string

import ray.rllib.env.multi_agent_env
import ray.rllib.utils.spaces
from ray import rllib
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
import numpy as np


from chicken_coop import county
from chicken_coop.county.constants import ALL_AGENTS
from .stating import State
from chicken_coop.coop_config import CoopConfig



class Env(county.BaseEnv):

    def __init__(self, config: Optional[Mapping] = None) -> None:
        county.BaseEnv.__init__(self, config)
        self.coop_config: CoopConfig = self.config.setdefault('coop_config', CoopConfig())

        self.agents = tuple(self.coop_config.policy_by_agent)
        self.policies = tuple(self.coop_config.policy_by_agent.values())

        self.action_space = gymnasium.spaces.Discrete(2)
        if self.coop_config.identity_is_linear:
            opponent_observation_space = gymnasium.spaces.Box(
                low=0,
                high=self.coop_config.n_agents - 1,
                shape=(2,),
                dtype=int,
            )
        else:
            opponent_observation_space = gymnasium.spaces.MultiBinary(self.coop_config.n_agents)

        observation_space_dict = {
            'i_opponent': opponent_observation_space,
        }

        if self.coop_config.episode_length >= 2:
            observation_space_dict |= {
                'i_turn': gymnasium.spaces.Box(
                    low=0,
                    high=self.coop_config.episode_length,
                    shape=(1,),
                    dtype=int,
                ),
                'n_victory_rewards': gymnasium.spaces.Box(
                    low=0,
                    high=self.coop_config.episode_length,
                    shape=(1,),
                    dtype=int,
                ),
                'n_boredom_rewards': gymnasium.spaces.Box(
                    low=0,
                    high=self.coop_config.episode_length,
                    shape=(1,),
                    dtype=int,
                ),
                'n_defeat_rewards': gymnasium.spaces.Box(
                    low=0,
                    high=self.coop_config.episode_length,
                    shape=(1,),
                    dtype=int,
                ),
                'n_anguish_rewards': gymnasium.spaces.Box(
                    low=0,
                    high=self.coop_config.episode_length,
                    shape=(1,),
                    dtype=int,
                ),
            }

        self.observation_space = gymnasium.spaces.Dict(observation_space_dict)

        self._agent_ids = set(self.agents)
        self.reset()


    def make_initial_state(self) -> State:
        return State.make_initial(coop_config=self.coop_config)

    @staticmethod
    def sample_episode_to_text(algorithm: ray.rllib.algorithms.Algorithm,
                               coop_config: CoopConfig) -> str:
        states = tuple(Env.sample_episode(algorithm))
        return '\n'.join(state.text for state in states)



from __future__ import annotations

from typing import Mapping
import abc

from chicken_coop.county.typing import Agent, Action, AgentOrAll, Observation, Reward, RealNumber


class BaseState(abc.ABC):

    observation_by_agent: Mapping[Agent, Observation]

    reward_by_agent: Mapping[Agent, Reward]

    terminated_by_agent: Mapping[AgentOrAll, bool]

    truncated_by_agent: Mapping[AgentOrAll, bool]

    text: str

    @abc.abstractstaticmethod
    def make_initial() -> BaseState: # In the future: `-> Self:`
        raise NotImplementedError

    @abc.abstractmethod
    def step(self, actions: Mapping[Agent, Action]) -> BaseState: # In the future: `-> Self:`
        raise NotImplementedError






from __future__ import annotations

import random
import re
import collections
import statistics
import tempfile
import time
import math
import os
import logging
import pathlib
import sys
import itertools
import warnings
import functools
from typing import Optional
import pathlib
import webbrowser

import click
import more_itertools
import pandas as pd
import numpy as np
import scipy.signal
import networkx as nx
import plotly.graph_objects as go
import plotly.colors
import plotly.subplots

import chicken_coop

from ..command_group import cli
from chicken_coop.county import misc
from chicken_coop.county import trekking
from chicken_coop.county.typing import RealNumber
from .. import utils
from ..dominance_hierarching import (DominanceHierarchy,
                                     AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy)
from . import constants
from . import plotting_dominance_hierarchies


ANIMAL_DATA_FOLDER = (pathlib.Path(__file__) / '../../../animal_data').resolve()
N_GENERATIONS_SHORT = 100
N_GENERATIONS_LONG = 150
Y_RANGE = (-0.05, 1.05)
BLUE_TO_RED = [
    [0, "rgb(0,0,255)"],
    [1, "rgb(255,0,0)"],
]

@cli.group()
@click.option('-p', '--path', 'trek_path_expression', type=str, required=False)
@click.option('-s', '--show/--dont-show', 'show_plot', is_flag=True, default=True)
@click.option('-w', '--write/--dont-write', 'write_plot', is_flag=True, default=True)
@click.option('-f', '--forgive-lack-of-samples', 'forgive_lack_of_samples', is_flag=True)
@click.pass_context
def plot_paper(context: click.Context, *, trek_path_expression: Optional[str],
               show_plot: bool, write_plot: bool,
               forgive_lack_of_samples: bool) -> None:
    context.ensure_object(dict)
    context.obj['trek_path_expression'] = trek_path_expression
    context.obj['show_plot'] = show_plot
    context.obj['write_plot'] = write_plot
    context.obj['forgive_lack_of_samples'] = forgive_lack_of_samples

global_layout_template = dict(
    plot_bgcolor='rgb(255, 255, 255, 0)',
    title_font_size=1,
    width=1_850,
    height=950,
)

global_axis_template = dict(
    showline=True,
    linewidth=3,
    linecolor='black',
)

global_y_axis_template = dict(
    range=Y_RANGE,
    tickformat='.1f',
)

def _show_and_or_write_plot(figure: go.Figure, *, name: str, context: click.Context) -> None:
    if context.obj['show_plot']:
        figure.show()
    if context.obj['write_plot']:
        folder: pathlib.Path = pathlib.Path.home() / 'Desktop'
        folder.mkdir(parents=True, exist_ok=True)
        figure.write_image(folder / f'{name}.png')



@plot_paper.command()
@click.pass_context
def comb_hero(context: click.Context) -> None:

    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False)

    print(f'Making a `comb-hero` plot for {trek} ...', file=sys.stderr)

    mini_treks = [mini_trek for mini_trek in trek.mini_treks
                  if (mini_trek.dominance_hierarchy.is_transitive and
                      mini_trek.dominance_hierarchy.is_complete)]

    dfs = [mini_trek.rankized_rollout_dataframe.filter(regex=r'rank\.[0-9]+\.aggressiveness')
           for mini_trek in mini_treks
           if mini_trek.rollout_dataframe.shape[0] >= N_GENERATIONS_SHORT + 1]
    df: pd.DataFrame = sum(dfs) / len(dfs)
    aggressiveness_column_names = tuple(name for name in df.columns
                                        if 'aggressiveness' in name and 'rank' in name)
    axis_template = global_axis_template | dict(
        title_font=dict(size=50),
        tickfont=dict(
            size=40,
            family='Noto Mono',
        ),
    )

    def make_label(column_name: str) -> str:
        i_rank = re.fullmatch(r'^rank\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Rank {i_rank}'

    figure = plotly.subplots.make_subplots(rows=2, cols=1)


    figure.add_traces(
        _make_chase_scatters(),
        rows=1,
        cols=1,
    )

    figure.update_xaxes(
        **axis_template,
        title_text='Interaction',
        row=1, col=1,
    )
    figure.update_yaxes(
        **axis_template,
        **global_y_axis_template,
        title_text='Aggressiveness',
        row=1, col=1,
    )

    figure.add_traces(
        tuple(
            go.Scatter(
                x=df.index, y=df[aggressiveness_column_name],
                name=make_label(aggressiveness_column_name),
                line=dict(
                    width=5.5,
                    color=plotly.colors.sample_colorscale(
                        BLUE_TO_RED,
                        i / (len(aggressiveness_column_names) - 1),
                        )[0]
                ),
            )
            for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
        ),
        rows=2,
        cols=1,
    )

    figure.update_xaxes(
        **axis_template,
        title_text='Generation',
        range=(0, N_GENERATIONS_SHORT),
        row=2, col=1,
    )
    figure.update_yaxes(
        **axis_template,
        **global_y_axis_template,
        title_text='Aggressiveness',
        row=2, col=1,
    )


    figure.update_layout(
        **global_layout_template | dict(
            height=2_000,
            showlegend=False,
        )
    )

    # for i in figure['layout']['annotations']:
        # i['font'] = dict(size=50)

    _show_and_or_write_plot(figure, name='comb_hero', context=context)


def _make_chase_scatters():
    N_AGENTS = 4
    ATTACKER = 'Attacker'
    VICTIM = 'Victim'
    ATTACK_TYPE = 'Attack Type'
    WINDOW_SIZE = 10

    chase_folder = ANIMAL_DATA_FOLDER / 'Chase2022'
    assert chase_folder.is_dir()

    output_dfs = []

    for csv_path in chase_folder.glob('*.csv'):
        input_df = pd.read_csv(csv_path, usecols=[ATTACKER, VICTIM, ATTACK_TYPE])
        input_df = input_df[input_df[ATTACK_TYPE] != 'J'].drop(columns=ATTACK_TYPE)

        fluffs = collections.defaultdict(list)
        gruffs = []

        for _, row in input_df.iterrows():
            fluffs[row[ATTACKER]].append(1)
            fluffs[row[VICTIM]].append(0)
            gruffs.append([misc.mean_default(items) if (items := fluffs[i][-WINDOW_SIZE:]) else None
                           for i in range(1, N_AGENTS + 1)])

        while None in gruffs[0]:
            del gruffs[:1]

        wip_df = pd.DataFrame(gruffs, columns=[f'agent_{i}_aggressiveness'
                                               for i in range(1, N_AGENTS + 1)])
        foo = dict(wip_df.mean())
        column_by_rank = tuple(sorted(foo, key=foo.__getitem__, reverse=True))
        output_df = pd.DataFrame({f'rank.{i}.aggressiveness': wip_df[column]
                                  for i, column in enumerate(column_by_rank)})
        output_dfs.append(output_df)

    final_df = sum(output_dfs)[:240] / len(output_dfs)

    aggressiveness_column_names = tuple(name for name in final_df.columns
                                        if 'aggressiveness' in name and 'rank' in name)

    def make_label(column_name: str) -> str:
        i_rank = re.fullmatch(r'^rank\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Rank {i_rank}'

    chase_scatters = tuple(
            go.Scatter(
                x=final_df.index, y=final_df[aggressiveness_column_name],
                name=make_label(aggressiveness_column_name),
                line=dict(
                    width=5.5,
                    color=plotly.colors.sample_colorscale(
                        BLUE_TO_RED,
                        i / (len(aggressiveness_column_names) - 1),
                    )[0]
                ),
            )
            for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
        )
    return chase_scatters


@plot_paper.command()
@click.pass_context
def samples_aggressiveness(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False)

    forgive_lack_of_samples = context.obj['forgive_lack_of_samples']

    print(f'Making a `samples-aggressiveness` plot for {trek} ...', file=sys.stderr)

    dfs = [mini_trek.rollout_dataframe.filter(regex=r'agent\.[0-9]+\.aggressiveness')
           for mini_trek in
           trek.get_sample_mini_treks(forgive_lack_of_samples=forgive_lack_of_samples)]
    aggressiveness_column_names = tuple(name for name in dfs[0].columns
                                        if 'aggressiveness' in name and 'agent' in name)
    axis_template = global_axis_template | dict(
        linewidth=2,
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    def make_label(column_name: str) -> str:
        i_agent = re.fullmatch(r'^agent\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Agent {i_agent}'

    subplots = []

    for df in dfs:
        subplots.append(
            tuple(
                go.Scatter(x=df.index, y=df[aggressiveness_column_name],
                           name=make_label(aggressiveness_column_name),
                           line=dict(width=4.5, color=plotly.colors.DEFAULT_PLOTLY_COLORS[i]))
                for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
            )
        )
    figure = plotly.subplots.make_subplots(
        rows=2, cols=2, shared_xaxes=True, shared_yaxes=True,
        horizontal_spacing=0.035,
        vertical_spacing=0.05,
        x_title='Generation',
        y_title='Aggressiveness',
    )

    axis_title_font_size = 55
    x_title, y_title = figure['layout']['annotations']
    x_title['font'] = y_title['font'] = dict(size=axis_title_font_size)
    x_title['yshift'] -= 10
    y_title['xshift'] -= 40

    for i, subplot in enumerate(subplots):
        figure.add_traces(subplot, rows=(1, 2)[i // 2], cols=[1, 2][i % 2])

    figure.update_layout(
        **global_layout_template,
        showlegend=False,
        margin=dict(l=140, b=100, t=10),
    )
    figure.update_xaxes(**axis_template, range=(0, N_GENERATIONS_SHORT))
    figure.update_yaxes(**axis_template, **global_y_axis_template)

    _show_and_or_write_plot(figure, name='samples_aggressiveness', context=context)


@plot_paper.command()
@click.pass_context
def samples_rapport(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False)

    forgive_lack_of_samples = context.obj['forgive_lack_of_samples']

    print(f'Making a `samples-rapport` plot for {trek} ...', file=sys.stderr)

    column_pattern = re.compile(r'^dialog\.([0-9]+)\.([0-9]+)\.rapport$')
    dfs = [mini_trek.rollout_dataframe.filter(regex=column_pattern.pattern)
           for mini_trek in
           trek.get_sample_mini_treks(forgive_lack_of_samples=forgive_lack_of_samples)]

    minimum_dfs = [pd.DataFrame({'Minimum rapport': df.min(axis=1)}) for df in dfs]

    axis_template = global_axis_template | dict(
        linewidth=2,
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    figure = plotly.subplots.make_subplots(
        rows=2, cols=2, shared_xaxes=True, shared_yaxes=True,
        horizontal_spacing=0.035,
        vertical_spacing=0.05,
        x_title='Generation',
        y_title='Minimum rapport',
    )

    axis_title_font_size = 55
    x_title, y_title = figure['layout']['annotations']
    x_title['font'] = y_title['font'] = dict(size=axis_title_font_size)
    x_title['yshift'] -= 10
    y_title['xshift'] -= 40

    for i, minimum_df in enumerate(minimum_dfs):
        figure.add_traces(
            go.Scatter(x=minimum_df.index, y=minimum_df.iloc[:, 0],
                       name=minimum_df.columns[0],
                       line=dict(width=4.5, color='black')),
            rows=(1, 2)[i // 2],
            cols=[1, 2][i % 2]
        )

    figure.update_layout(
        **global_layout_template,
        showlegend=False,
        margin=dict(l=140, b=100, t=10),
    )
    figure.update_xaxes(**axis_template)
    figure.update_yaxes(**axis_template, **global_y_axis_template)

    _show_and_or_write_plot(figure, name='samples_rapport', context=context)


@plot_paper.command()
@click.pass_context
def samples_dh(context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False)

    forgive_lack_of_samples = context.obj['forgive_lack_of_samples']

    print(f'Making a `samples-dh` plot for {trek} ...', file=sys.stderr)

    dominance_hierarchies = [
        mini_trek.dominance_hierarchy for mini_trek in
        trek.get_sample_mini_treks(forgive_lack_of_samples=forgive_lack_of_samples)
    ]

    plot_svg = plotting_dominance_hierarchies.make_condensed_plot_svg(dominance_hierarchies)

    # dominance_hierarchy = DominanceHierarchy(
        # # '(1 -> 2 -> 3 -> 1) => 4 => 0 => (6 -> 7 -> 8 -> 9 -> 6 -> 8 | 9 -> 7) => 5'
        # # '(0 -> 1 -> 2 -> 3 -> 4 -> 0 | 2 -> 0 -> 3 -> 1 -> 4 -> 2) => 5'
        # # '0 => 1 => 2 => 3 => 4 => 5 => 6'
        # # '0 => 1 => (2 -> 3 -> 4 -> 2) => 5 => 6'
        # # '0 => (1 -> 2 -> 3 -> 4 -> 1 | 1 -> 3 | 2 -> 4) => 5 => 6'
        # '0 => (1 -> 2 -> 3 -> 1) => (4 -> 5 -> 6 -> 4)'
    # )

    indent = ' ' * 8
    content = constants.HTML_TEMPLATE.format(indent + plot_svg.replace('\n', '\n' + indent))
    if context.obj['write_plot']:
        (pathlib.Path.home() / 'Desktop' / 'samples_dh.html').write_text(content)
    if context.obj['show_plot']:
        with tempfile.NamedTemporaryFile('w', suffix='.html') as fp:
            fp.write(content)
            fp.flush()
            webbrowser.open_new_tab(fp.name)
            time.sleep(3)


@plot_paper.command()
@click.pass_context
def ablate_observation_rapport(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False,
                             moniker='paper-ablate-observation')


    print(f'Making an `ablate-observation-rapport` plot for {trek} ...', file=sys.stderr)
    mini_treks_by_observation_accuracy = trek.get_mini_treks_grouped_by(
        lambda mini_trek: mini_trek.meta['coop_config']['observation_accuracy']
    )
    assert set(mini_treks_by_observation_accuracy) == {0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                                                       0.9, 1.0}

    df_by_observation_accuracy = {}

    for observation_accuracy, mini_treks in mini_treks_by_observation_accuracy.items():
        serieses = [mini_trek.rollout_dataframe['population.rapport'] for mini_trek in mini_treks
                    if mini_trek.rollout_dataframe.shape[0] >= N_GENERATIONS_LONG + 1]
        df_by_observation_accuracy[observation_accuracy] = pd.concat(serieses, axis=1
                                                             ).mean(axis=1).to_frame(name='rapport')

    axis_template = global_axis_template | dict(
        title_font=dict(size=50),
        tickfont=dict(
            size=40,
            family='Noto Mono',
        ),
    )

    figure = go.Figure(
        data=tuple(
            go.Scatter(x=df.index, y=df['rapport'],
                       name=str(observation_accuracy),
                       line=dict(
                           width=5.5,
                           color=plotly.colors.sample_colorscale(BLUE_TO_RED,
                                                                 [observation_accuracy])[0]
                       ))
            for observation_accuracy, df in df_by_observation_accuracy.items()
        ),
        layout=go.Layout(
            **global_layout_template,
            xaxis=dict(
              **axis_template,
              title='Generation',
              range=(0, N_GENERATIONS_LONG),
            ),
            yaxis=dict(
              **axis_template,
              **global_y_axis_template,
              title=dict(text='Rapport', standoff=0),
            ),
            legend=dict(
                xanchor='left', yanchor='top',
                x=-0.25, y=1,
                traceorder='reversed',
                font=dict(size=40)
                ),
            annotations=(
                go.layout.Annotation(
                    text='Opponent perception accuracy',
                    xref='paper', yref='paper',
                    # x=-0.79, y=0.5,
                    x=-0.27, y=0.5,
                    xanchor='center',
                    yanchor='middle',
                    textangle=-90,
                    font=dict(
                        size=50,
                    ),
                ),
            ),
            margin=dict(l=400),
        ),
    )
    _show_and_or_write_plot(figure, name='ablate_observation_rapport', context=context)


@plot_paper.command()
@click.option('-o', '--output', is_flag=True)
@click.pass_context
def ablate_observation_samples_aggressiveness(context, *, output: bool) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False,
                             moniker='paper-ablate-observation')

    dfs = [mini_trek.rollout_dataframe.filter(regex=r'agent\.[0-9]+\.aggressiveness')
           for mini_trek in
           trek.get_mini_treks_grouped_by(
               lambda mini_trek: mini_trek.meta['coop_config']['observation_accuracy'])[0.1][:4]]
    assert len(dfs) == 4

    aggressiveness_column_names = tuple(name for name in dfs[0].columns
                                        if 'aggressiveness' in name and 'agent' in name)
    axis_template = global_axis_template | dict(
        linewidth=2,
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    def make_label(column_name: str) -> str:
        i_agent = re.fullmatch(r'^agent\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Agent {i_agent}'

    subplots = []

    for df in dfs:
        subplots.append(
            tuple(
                go.Scatter(x=df.index, y=df[aggressiveness_column_name],
                           name=make_label(aggressiveness_column_name),
                           line=dict(width=4.5, color=plotly.colors.DEFAULT_PLOTLY_COLORS[i]))
                for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
            )
        )
    figure = plotly.subplots.make_subplots(
        rows=2, cols=2, shared_xaxes=True, shared_yaxes=True,
        horizontal_spacing=0.035,
        vertical_spacing=0.05,
        x_title='Generation',
        y_title='Aggressiveness',
    )

    axis_title_font_size = 55
    x_title, y_title = figure['layout']['annotations']
    x_title['font'] = y_title['font'] = dict(size=axis_title_font_size)
    x_title['yshift'] -= 10
    y_title['xshift'] -= 40

    for i, subplot in enumerate(subplots):
        figure.add_traces(subplot, rows=(1, 2)[i // 2], cols=[1, 2][i % 2])

    figure.update_layout(
        **global_layout_template,
        showlegend=False,
        margin=dict(l=140, b=100, t=10),
    )
    figure.update_xaxes(**axis_template, range=(0, N_GENERATIONS_SHORT))
    figure.update_yaxes(**axis_template, **global_y_axis_template)

    _show_and_or_write_plot(figure, name='ablate_observation_samples_aggressiveness',
                            context=context)


@plot_paper.command()
@click.pass_context
def transmit(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False,
                             moniker='paper-transplant')

    print(f'Making a `transmit` plot for {trek} ...', file=sys.stderr)
    mini_treks_by_n_visitor_agents = trek.get_mini_treks_grouped_by(
        lambda mini_trek: len(mini_trek.meta['coop_config']['i_visitor_agents'])
    )

    df_by_n_visitor_agents = {}

    for n_visitor_agents, mini_treks in sorted(mini_treks_by_n_visitor_agents.items()):
        df_by_n_visitor_agents[n_visitor_agents] = pd.DataFrame(
            [1 - mini_trek.rollout_dataframe['population.'
                             'restricted_distance_from_visitor_dominance_hierarchy'].tail(1).iloc[0]
             for mini_trek in mini_treks],
            columns=('transmission_fidelidy',),
        )

    axis_template = global_axis_template | dict(
        title_font=dict(size=50),
        tickfont=dict(
            size=40,
            family='Noto Mono',
        ),
    )

    figure = go.Figure(
        data=tuple(
            go.Box(
                x=[n_visitor_agents] * len(df['transmission_fidelidy']),
                y=df['transmission_fidelidy'],
                name=str(n_visitor_agents),
                marker=dict(
                    color='black',
                ),
                # notched=True,
                showwhiskers=False,
                boxpoints=False,
                quartilemethod='exclusive',
                # boxmean=True,
                line=dict(width=4,),
            ) for n_visitor_agents, df in df_by_n_visitor_agents.items()
        ) + tuple(
            go.Scatter(
                x=[n_visitor_agents],
                y=[df['transmission_fidelidy'].median()],
                name=str(n_visitor_agents),
                marker=dict(
                    color='white',
                    size=60,
                    line=dict(
                        width=4,
                        color='black',
                    )
                )
            ) for n_visitor_agents, df in df_by_n_visitor_agents.items()
        )
        ,
        layout=go.Layout(
            **global_layout_template,
            xaxis=dict(
              **axis_template,
              title='Number of experienced agents',
            ),
            yaxis=axis_template | global_y_axis_template | dict(
              title_font=dict(size=47),
              title=dict(text='Dominance hierarchy transmission fidelidy'),
            ),
            showlegend=False,
        ),
    )
    _show_and_or_write_plot(figure, name='transmit', context=context)


@plot_paper.command()
@click.pass_context
def non_linearity(context: click.Context) -> None:

    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False,
                             moniker='paper-cycles')

    print(f'Making a `non-linearity` plot for {trek} ...', file=sys.stderr)

    def mouse_string_to_i_agent(agent_string: str) -> int:
        i_agent_plus_one = int(re.fullmatch('^M([0-9]+)$', agent_string).group(1))
        assert agent_string == f'M{i_agent_plus_one}'
        return i_agent_plus_one - 1

    def csv_path_to_dominance_hierarchy(path: pathlib.Path) -> DominanceHierarchy:
        df = pd.read_csv(path, index_col=0)

        edges = []

        for i in range(df.shape[0]):
            for j in range(i+1, df.shape[1]): # We start from i+1 to avoid repeating pairs
                agent_a = df.index[i]
                agent_b = df.columns[j]

                a_wins = df.at[agent_a, agent_b]
                b_wins = df.at[agent_b, agent_a]

                if pd.isna(a_wins) or pd.isna(b_wins):
                    assert pd.isna(a_wins) and pd.isna(b_wins)
                    assert agent_a == agent_b
                    continue
                assert agent_a != agent_b

                if a_wins >= b_wins:
                    edges.append((mouse_string_to_i_agent(agent_a),
                                  mouse_string_to_i_agent(agent_b)))
                else:
                    assert b_wins > a_wins
                    edges.append((mouse_string_to_i_agent(agent_b),
                                  mouse_string_to_i_agent(agent_a)))

        return DominanceHierarchy(edges)

    williamson_folder = ANIMAL_DATA_FOLDER / 'Williamson2017'

    csv_paths = tuple(williamson_folder.glob('*.csv'))

    mice_dominance_hierarchies = tuple(map(csv_path_to_dominance_hierarchy, csv_paths))
    mice_mean_rank_legibilities = \
                             DominanceHierarchy.get_mean_rank_linearity(mice_dominance_hierarchies)


    dominance_hierarchies = tuple(mini_trek.dominance_hierarchy for mini_trek in trek.mini_treks
                                  if mini_trek.dominance_hierarchy.is_complete)

    mean_rank_legibilities = DominanceHierarchy.get_mean_rank_linearity(dominance_hierarchies)

    df = pd.DataFrame({'mice_rank_linearity': mice_mean_rank_legibilities,
                       'chicken_coop_rank_linearity': mean_rank_legibilities})

    axis_template = global_axis_template | dict(
        title_font=dict(size=50),
        tickfont=dict(
            size=40,
            family='Noto Mono',
        ),
    )

    figure = go.Figure(
        data=(
            go.Scatter(
                x=df.index,
                y=df['mice_rank_linearity'],
                name='Male CD-1 mice',
                mode='lines+markers',
                line=dict(
                    color='orange',
                    width=6,
                ),
                marker=dict(
                    size=13,
                )
            ),
            go.Scatter(
                x=df.index,
                y=df['chicken_coop_rank_linearity'],
                name='Chicken Coop agents',
                mode='lines+markers',
                line=dict(
                    color='purple',
                    width=6,
                ),
                marker=dict(
                    size=13,
                )
            ),
        )
        ,
        layout=go.Layout(
            **global_layout_template,
            xaxis=dict(
              **axis_template,
              title='Rank',
              dtick=1,
            ),
            yaxis=dict(
              **axis_template,
              **global_y_axis_template,
              title=dict(text='Rank linearity'),
            ),
            showlegend=True,
            legend=dict(
                xanchor='right', yanchor='top',
                font=dict(size=40)
            ),
        ),
    )
    _show_and_or_write_plot(figure, name='non_linearity', context=context)





import numbers
from typing import Any
from ray.rllib.utils.typing import EnvActionType as Action, PolicyID, MultiAgentDict

RealNumber = float | int | numbers.Real

Agent = str
AgentOrAll = str
Observation = Any
Reward = RealNumber

del Any, numbers


from __future__ import annotations

import itertools
import re
from typing import Any

import networkx as nx
import more_itertools


def is_oriented(di_graph: nx.DiGraph) -> bool:
    for node_a, node_b in itertools.combinations(di_graph, 2):
        if (((node_a, node_b) in di_graph.edges) and ((node_b, node_a) in di_graph.edges)):
            return False
    return True

def is_tournament(di_graph: nx.DiGraph) -> bool:
    return len(di_graph) <= 1 or ((nx.density(di_graph) == 0.5) and is_oriented(di_graph))


def scc_di_graph_to_concise_text(scc_di_graph: nx.DiGraph) -> str:
    assert nx.is_strongly_connected(scc_di_graph)
    assert is_tournament(scc_di_graph)
    assert all(isinstance(node, int) for node in scc_di_graph)
    if len(scc_di_graph) == 1:
        (node,) = scc_di_graph
        return str(node)
    else:
        available_edges = sorted(scc_di_graph.edges)
        fluffs = []
        while available_edges:
            fluffs.append(fluff := [available_edges.pop(0)])
            while True:
                for edge in available_edges:
                    if edge[0] == fluff[-1][1]:
                        available_edges.remove(edge)
                        fluff.append(edge)
                        break
                else:
                    break

        fluff_texts = [
            ' -> '.join(map(str, (fluff[0][0],) + tuple(zip(*fluff, strict=True))[1]))
            for fluff in fluffs
        ]
        return '(' + ' | '.join(fluff_texts) + ')'


def oriented_graph_to_concise_text(oriented_graph: nx.DiGraph) -> str:
    assert is_oriented(oriented_graph)
    assert all(isinstance(node, int) for node in oriented_graph)
    oriented_graph_without_isolated_nodes = nx.DiGraph(oriented_graph.edges)

    if is_tournament(oriented_graph_without_isolated_nodes):
        condensed_di_graph = nx.condensation(oriented_graph_without_isolated_nodes)
        scc_concise_texts = []
        for meta_node in nx.topological_sort(condensed_di_graph):
            nodes = set(condensed_di_graph.nodes.data()[meta_node]['members'])
            scc_di_graph = get_sub_di_graph(oriented_graph_without_isolated_nodes, nodes)
            scc_concise_texts.append(scc_di_graph_to_concise_text(scc_di_graph))
        result = ' => '.join(scc_concise_texts)
    else:
        result = ', '.join(f'({edge[0]}, {edge[1]})' for edge in oriented_graph.edges())

    isolated_nodes = tuple(nx.isolates(oriented_graph))
    if isolated_nodes:
        inner = ' '.join(f'{isolated_node}?' for isolated_node in isolated_nodes)
        result += f' <{inner}>'
    return result


_scc_di_graph_subpattern = re.compile(
    '(?:([1-9][0-9]*) -> )*([1-9][0-9]*)'
)
_scc_di_graph_pattern = re.compile(
    fr' \| '
)

def concise_text_to_scc_di_graph(concise_text: str) -> nx.DiGraph:
    if concise_text.isnumeric():
        di_graph = nx.DiGraph({int(concise_text): ()})
    else:
        inner_text = re.fullmatch(r'^\(([0-9| \->]+)\)$', concise_text).group(1)
        parts = inner_text.split(' | ')
        di_graph = nx.DiGraph(
            itertools.chain.from_iterable(
                more_itertools.windowed(map(int, part.split(' -> ')), 2) for part in parts
            )
        )
    assert nx.is_strongly_connected(di_graph)
    assert is_tournament(di_graph)
    return di_graph


isolated_nodes_pattern = re.compile(r' <(?:[0-9]+\? )*[0-9]+\?>$')
isolated_nodes_subpattern = re.compile(r'^$')

def concise_text_to_oriented_graph(concise_text: str) -> nx.DiGraph:
    isolated_nodes_match = isolated_nodes_pattern.search(concise_text)
    isolated_nodes = (() if isolated_nodes_match is None
                      else tuple(map(int, re.findall('[0-9]+', isolated_nodes_match.group(0)))))
    short_concise_text = isolated_nodes_pattern.sub('', concise_text)
    if '=>' not in short_concise_text and '->' not in short_concise_text:
        di_graph = nx.DiGraph(eval(f'({short_concise_text})')) # Ugly hack
    else:
        scc_di_graphs = tuple(map(concise_text_to_scc_di_graph, short_concise_text.split(' => ')))
        di_graph = nx.DiGraph(
            itertools.chain.from_iterable(
                itertools.chain(
                    (scc_di_graph.edges for scc_di_graph in scc_di_graphs),
                    itertools.starmap(itertools.product, itertools.combinations(scc_di_graphs, 2))
                )
            )
        )
    assert not (set(isolated_nodes) & set(di_graph))
    di_graph.add_nodes_from(isolated_nodes)
    assert is_oriented(di_graph)
    return di_graph


def get_sub_di_graph(di_graph: nx.DiGraph, nodes: Iterable) -> nx.DiGraph:
    node_set = set(nodes)
    sub_di_graph = nx.DiGraph(edge for edge in di_graph.edges if set(edge) <= node_set)
    sub_di_graph.add_nodes_from(nodes)
    return sub_di_graph





from __future__ import annotations

import random
import pprint
import re
import collections
import statistics
import tempfile
import time
import math
import os
import logging
import pathlib
import sys
import itertools
import warnings
import functools
from typing import Optional
import pathlib
import webbrowser

import click
import yaml
import more_itertools
import pandas as pd
import numpy as np
import scipy.signal
import networkx as nx
import plotly.graph_objects as go
import plotly.colors
import plotly.subplots

import chicken_coop

from ..command_group import cli
from chicken_coop.county import misc
from chicken_coop.county import trekking
from chicken_coop.county.typing import RealNumber
from .. import utils
from ..dominance_hierarching import (DominanceHierarchy,
                                     AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy)
from . import constants
from . import plotting_dominance_hierarchies



@cli.group()
@click.option('-p', '--path', 'trek_path_expression', type=str, required=False)
@click.option('-f', '--forgive-lack-of-samples', 'forgive_lack_of_samples', is_flag=True)
@click.pass_context
def analyze_paper(context: click.Context, *, trek_path_expression: Optional[str],
                  forgive_lack_of_samples: bool) -> None:
    context.ensure_object(dict)
    context.obj['trek_path_expression'] = trek_path_expression
    context.obj['forgive_lack_of_samples'] = forgive_lack_of_samples




@analyze_paper.command()
@click.pass_context
def analyzo(context: click.Context) -> None:

    trek = trekking.Trek.get(context.obj['trek_path_expression'], allow_single_trek=False)

    print(f'Making an `analyzo` analysis for {trek} ...', file=sys.stderr)

    mini_treks = [mini_trek for mini_trek in trek.mini_treks
                  if mini_trek.dominance_hierarchy.is_complete]
    counter = collections.Counter(mini_trek.dominance_hierarchy for mini_trek in mini_treks)

    result = {
        'n_populations': len(mini_treks),
        'n_populations_that_converged_to_dominance_hierarchy': sum(
            mini_trek.dominance_hierarchy.is_complete for mini_trek in mini_treks
        ),
        'n_dominance_hierarchies': len(counter),
        'n_transitive_dominance_hierarchies': (
            n_transitive_dominance_hierarchies := sum(dominance_hierarchy.is_transitive
                                                      for dominance_hierarchy in counter)
        ),
        'n_intransitive_dominance_hierarchies': len(counter) - n_transitive_dominance_hierarchies,
        'transitive_ratio': n_transitive_dominance_hierarchies / len(counter),
        'most_common': list(list(zip(*counter.most_common(3)))[1]),
        'least_common': list(list(zip(*counter.most_common()[:-4:-1]))[1]),
    }
    print(yaml.dump(result))



from . import bootstrap

from .dominance_hierarching import DominanceHierarchy
from .coop_config import CoopConfig
from .stating import State
from .enving import Env

from .command_group import cli
from . import running
from . import plotting


__version__ = '0.0.2'


from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable
import abc
import statistics
import enum
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import functools
import dataclasses

import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
from frozendict import frozendict
import numpy as np
from scipy import stats
import more_itertools

from chicken_coop import county
from chicken_coop.county import misc
from chicken_coop.county.typing import Agent, RealNumber
from chicken_coop.county.constants import ALL_AGENTS


class CoopException(Exception):
    pass


DEFAULT_VICTORY_REWARD = 5
DEFAULT_BOREDOM_REWARD = 0
DEFAULT_DEFEAT_REWARD = -2
DEFAULT_ANGUISH_REWARD = -10

DEFAULT_EPISODE_LENGTH = 1
DEFAULT_N_GENERATIONS = 150
DEFAULT_N_TUNE_SAMPLES = 5
DEFAULT_TRAIN_BATCH_SIZE = 512 # 4_000
DEFAULT_N_AGENTS = 6
DEFAULT_IDENTITY_IS_LINEAR = False
DEFAULT_POLICY_SNAPSHOT_PERIOD = 100
DEFAULT_LEARNING_RATE = 2e-06
DEFAULT_N_ROLLOUT_WORKERS = 2
DEFAULT_FLIP_INITIAL_WEIGHTS = False
DEFAULT_DOMINANCE_RELATIONSHIP_POLARITY_THRESHOLD = 0.55
DEFAULT_OBSERVATION_ACCURACY = 1

DEFAULT_N_VISITOR_POPULATIONS_PER_N_VISITOR_AGENTS = 10
DEFAULT_N_TRANSPLANTS_PER_VISITOR_POPULATION = 30

DEFAULT_FREEZE_VISITORS = True


from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable
import threading
import pathlib
import time
import traceback as traceback_module
import sys
import datetime

class ThreadyKrueger(threading.Thread):
    _stop = False
    def __init__(self, path: pathlib.Path | str, *, start: bool = False,
                 interval_seconds: int = 60) -> None:
        self.path = pathlib.Path(path)
        self.interval_seconds = interval_seconds
        if self.path.exists():
            assert self.path.is_dir()
        threading.Thread.__init__(self, daemon=True)
        if start:
            self.start()

    def run(self) -> None:
        self.path.mkdir(parents=True, exist_ok=True)
        while not self._stop:
            with (self.path / datetime.datetime.now().isoformat()).open('w') as f:
                for thread_id, frame in sys._current_frames().items():
                    f.write(f'Thread ID: {thread_id}\n')
                    f.write(f'{"".join(traceback_module.format_stack(frame))}\n')
            time.sleep(self.interval_seconds)

    def __del__(self) -> None:
        self.stop()

    def stop(self) -> None:
        self._stop = True






from __future__ import annotations

import random
import re
import collections
import statistics
import tempfile
import time
import math
import os
import logging
import pathlib
import sys
import itertools
import warnings
import functools
from typing import Optional
import pathlib
import webbrowser

import click
import more_itertools
import pandas as pd
import numpy as np
import scipy.signal
import networkx as nx
import plotly.graph_objects as go
import plotly.colors
import plotly.subplots

import chicken_coop

from ..command_group import cli
from chicken_coop.county import misc
from chicken_coop.county import trekking
from chicken_coop.county.typing import RealNumber
from .. import utils
from ..dominance_hierarching import (DominanceHierarchy,
                                     AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy)
from . import constants
from . import plotting_dominance_hierarchies


ANIMAL_DATA_FOLDER = (pathlib.Path(__file__) / '../../../animal_data').resolve()
N_GENERATIONS_SHORT = 100
N_GENERATIONS_LONG = 150
Y_RANGE = (-0.05, 1.05)
BLUE_TO_RED = [
    [0, "rgb(0,0,255)"],
    [1, "rgb(255,0,0)"],
]

@cli.group()
@click.option('-p', '--path', 'trek_path_expression', type=str, required=False)
@click.option('-s', '--show/--dont-show', 'show_plot', is_flag=True, default=True)
@click.option('-w', '--write/--dont-write', 'write_plot', is_flag=True, default=True)
@click.option('-f', '--forgive-lack-of-samples', 'forgive_lack_of_samples', is_flag=True)
@click.pass_context
def plot_deck(context: click.Context, *, trek_path_expression: Optional[str],
              show_plot: bool, write_plot: bool,
              forgive_lack_of_samples: bool) -> None:
    context.ensure_object(dict)
    context.obj['trek_path_expression'] = trek_path_expression
    context.obj['show_plot'] = show_plot
    context.obj['write_plot'] = write_plot
    context.obj['forgive_lack_of_samples'] = forgive_lack_of_samples

global_layout_template = dict(
    plot_bgcolor='rgb(255, 255, 255, 0)',
    title_font_size=1,
    width=1_850,
    height=950,
)

global_axis_template = dict(
    showline=True,
    linewidth=3,
    linecolor='black',
)

global_y_axis_template = dict(
    range=Y_RANGE,
    tickformat='.1f',
)

def _show_and_or_write_plot(figure: go.Figure, *, name: str, context: click.Context) -> None:
    if context.obj['show_plot']:
        figure.show()
    if context.obj['write_plot']:
        folder: pathlib.Path = pathlib.Path.home() / 'Desktop'
        folder.mkdir(parents=True, exist_ok=True)
        figure.write_image(folder / f'{name}.png')



@plot_deck.command()
@click.pass_context
def two_aggressiveness(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], moniker='deck-two',
                             allow_single_trek=False)

    print(f'Making `two-aggressiveness` plots for {trek} ...', file=sys.stderr)

    dfs = [mini_trek.rollout_dataframe.filter(regex=r'agent\.[0-9]+\.aggressiveness')
           for mini_trek in trek.mini_treks[:10]]
    aggressiveness_column_names = tuple(sorted(name for name in dfs[0].columns
                                               if 'aggressiveness' in name and 'agent' in name))
    axis_template = global_axis_template | dict(
        linewidth=2,
        title_font=dict(size=50),
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    def make_label(column_name: str) -> str:
        i_agent = re.fullmatch(r'^agent\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Agent {i_agent}'


    for i, df in enumerate(dfs):
        figure = go.Figure(
            data=tuple(
                go.Scatter(
                    x=df.index, y=df[aggressiveness_column_name],
                    name=make_label(aggressiveness_column_name),
                    line=dict(
                        width=5.5,
                        color=plotly.colors.sample_colorscale(
                            BLUE_TO_RED,
                            i / (len(aggressiveness_column_names) - 1),
                            )[0]
                    ),
                )
                for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
            ),
            layout=go.Layout(
                **global_layout_template,
                xaxis=dict(
                  **axis_template,
                  title='Generation',
                  range=(0, N_GENERATIONS_SHORT),
                ),
                yaxis=dict(
                  **axis_template,
                  **global_y_axis_template,
                  title=dict(text='Aggressiveness'),
                ),
                showlegend=True,
                legend=dict(
                    xanchor='right', yanchor='middle',
                    y=0.5,
                    font=dict(size=40)
                ),
            ),
        )

        _show_and_or_write_plot(figure, name=f'two_aggressiveness_{i}', context=context)


@plot_deck.command()
@click.pass_context
def six_aggressiveness(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], moniker='deck-six',
                             allow_single_trek=False)

    print(f'Making `six-aggressiveness` plots for {trek} ...', file=sys.stderr)

    dfs = [mini_trek.rollout_dataframe.filter(regex=r'agent\.[0-9]+\.aggressiveness')
           for mini_trek in trek.mini_treks[:10]]
    aggressiveness_column_names = tuple(sorted(name for name in dfs[0].columns
                                               if 'aggressiveness' in name and 'agent' in name))
    axis_template = global_axis_template | dict(
        linewidth=2,
        title_font=dict(size=50),
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    def make_label(column_name: str) -> str:
        i_agent = re.fullmatch(r'^agent\.([0-9]+)\.aggressiveness$', column_name).group(1)
        return f'Agent {i_agent}'


    for i, df in enumerate(dfs):
        figure = go.Figure(
            data=tuple(
                go.Scatter(
                    x=df.index, y=df[aggressiveness_column_name],
                    name=make_label(aggressiveness_column_name),
                    line=dict(
                        width=5.5,
                        color=plotly.colors.DEFAULT_PLOTLY_COLORS[i]
                    ),
                )
                for i, aggressiveness_column_name in enumerate(aggressiveness_column_names)
            ),
            layout=go.Layout(
                **global_layout_template,
                xaxis=dict(
                  **axis_template,
                  title='Generation',
                  range=(0, N_GENERATIONS_SHORT),
                ),
                yaxis=dict(
                  **axis_template,
                  **global_y_axis_template,
                  title=dict(text='Aggressiveness'),
                ),
                showlegend=True,
                legend=dict(
                    xanchor='left', yanchor='middle',
                    y=0.8,
                    x=0.02,
                    font=dict(size=32),
                    traceorder='normal',
                ),
            ),
        )

        _show_and_or_write_plot(figure, name=f'six_aggressiveness_{i}', context=context)


@plot_deck.command()
@click.pass_context
def six_aggressiveness_pair(context: click.Context) -> None:
    trek = trekking.Trek.get(context.obj['trek_path_expression'], moniker='deck-six',
                             allow_single_trek=False)

    i_agent_pair = (1, 4)

    print(f'Making `six-aggressiveness-pair` plots for {trek} ...', file=sys.stderr)

    dfs = [mini_trek.rollout_dataframe.filter(
                regex=fr'dialog\.{i_agent_pair[0]}\.{i_agent_pair[1]}\.(left|right)_aggressiveness')
           for mini_trek in trek.mini_treks[:10]]
    aggressiveness_column_names = tuple(sorted(dfs[0].columns))
    axis_template = global_axis_template | dict(
        linewidth=2,
        title_font=dict(size=50),
        tickfont=dict(
            size=35,
            family='Noto Mono',
        ),
    )

    for i, df in enumerate(dfs):
        figure = go.Figure(
            data=tuple(
                go.Scatter(
                    x=df.index, y=df[aggressiveness_column_name],
                    name=f'Agent {i_agent}',
                    line=dict(
                        width=5.5,
                        color=plotly.colors.DEFAULT_PLOTLY_COLORS[i_agent]
                    ),
                )
                for i_agent, aggressiveness_column_name in zip(i_agent_pair,
                                                               aggressiveness_column_names)
            ),
            layout=go.Layout(
                **global_layout_template,
                xaxis=dict(
                  **axis_template,
                  title='Generation',
                  range=(0, N_GENERATIONS_SHORT),
                ),
                yaxis=dict(
                  **axis_template,
                  **global_y_axis_template,
                  title=dict(text='Aggressiveness'),
                ),
                showlegend=True,
                legend=dict(
                    xanchor='right', yanchor='middle',
                    y=0.5,
                    font=dict(size=40)
                ),
            ),
        )

        _show_and_or_write_plot(figure, name=f'six_aggressiveness_pair_{i}', context=context)


@plot_deck.command()
@click.argument('mini_trek_path_string')
@click.pass_context
def dh(context, *, mini_trek_path_string: str) -> None:
    assert not context.obj['trek_path_expression']
    mini_trek_path = pathlib.Path(mini_trek_path_string)
    assert mini_trek_path.is_dir()
    mini_trek = trekking.MiniTrek(mini_trek_path)

    print(f'Making a `dh` plot for {mini_trek} ...', file=sys.stderr)

    plot_svg = plotting_dominance_hierarchies.make_condensed_plot_svg(mini_trek.dominance_hierarchy)

    indent = ' ' * 8
    content = constants.HTML_TEMPLATE.format(indent + plot_svg.replace('\n', '\n' + indent))
    if context.obj['write_plot']:
        (pathlib.Path.home() / 'Desktop' / 'dh.html').write_text(content)
    if context.obj['show_plot']:
        with tempfile.NamedTemporaryFile('w', suffix='.html') as fp:
            fp.write(content)
            fp.flush()
            webbrowser.open_new_tab(fp.name)
            time.sleep(3)





from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, TypeVar
import abc
import sys
import hashlib
import statistics
import enum
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import functools
import dataclasses

import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
from frozendict import frozendict
import numpy as np
from scipy import stats
import more_itertools

from chicken_coop.coop_config import CoopConfig
from .zeitgeisting import Zeitgeist, Dialog
from chicken_coop import county
from chicken_coop.county import misc
from chicken_coop.county.typing import Agent, RealNumber
from chicken_coop.county.constants import ALL_AGENTS


class Move(enum.Enum):
    DOVE = 'dove'
    HAWK = 'hawk'

    @staticmethod
    def from_neural(neural: int) -> Move:
        if neural == 0:
            return Move.DOVE
        else:
            assert neural == 1
            return Move.HAWK

    def to_neural(self) -> int:
        if self == Move.DOVE:
            return 0
        else:
            assert self == Move.HAWK
            return 1

    @staticmethod
    def get_random() -> Move:
        return random.choice((Move.HAWK, Move.DOVE))

    @property
    def as_letter(self) -> str:
        return self.value[0]



@dataclasses.dataclass(frozen=True, kw_only=True)
class State(county.BaseState):

    coop_config: CoopConfig
    i_turn: int
    joint_moves: tuple[tuple[Move, Move], ...]
    joint_rewards: tuple[tuple[RealNumber, RealNumber], ...]
    i_agent_pairs: tuple[tuple[int, int], ...]
    episode_seed: int


    def __post_init__(self) -> None:
        assert set(itertools.chain.from_iterable(self.i_agent_pairs)) == set(self.i_agents)
        assert all(map(more_itertools.is_sorted, self.i_agent_pairs))

    @property
    def n_agents(self) -> int:
        return self.coop_config.n_agents

    @property
    def i_agents(self) -> tuple[int, ...]:
        return self.coop_config.i_agents


    @functools.cached_property
    def agents(self) -> tuple[Agent, ...]:
        return tuple(self.coop_config.policy_by_agent)



    @functools.cached_property
    def observation_by_agent(self) -> dict[Agent, dict]:
        observation_by_agent = {}
        for i_agent, agent in enumerate(self.agents):
            i_observed_opponent = self.i_observed_opponent_by_i_agent[i_agent]
            observation_by_agent[agent] = {
                'i_opponent': (
                    np.array(
                        [i_observed_opponent,
                         self.coop_config.n_agents - 1 - i_observed_opponent],
                        dtype=int
                    ) if self.coop_config.identity_is_linear else
                    misc.make_one_hot(i_observed_opponent, self.coop_config.n_agents)
                )
            }

            if self.coop_config.episode_length >= 2:
                observation_by_agent[agent] |= {
                    'i_turn': np.array([self.i_turn], dtype=int),
                    'n_victory_rewards': np.array(
                        [self.rewards_by_i_agent[i_agent].count(self.coop_config.victory_reward)],
                        dtype=int,
                    ),
                    'n_boredom_rewards': np.array(
                        [self.rewards_by_i_agent[i_agent].count(self.coop_config.boredom_reward)],
                        dtype=int,
                    ),
                    'n_defeat_rewards': np.array(
                        [self.rewards_by_i_agent[i_agent].count(self.coop_config.defeat_reward)],
                        dtype=int,
                    ),
                    'n_anguish_rewards': np.array(
                        [self.rewards_by_i_agent[i_agent].count(self.coop_config.anguish_reward)],
                        dtype=int,
                    ),
                }

        return observation_by_agent

    @functools.cached_property
    def i_opponent_by_i_agent(self) -> dict[int, int]:
        opponent_by_agent = {}
        for (agent_0, agent_1) in self.i_agent_pairs:
            opponent_by_agent[agent_0] = agent_1
            opponent_by_agent[agent_1] = agent_0
        return opponent_by_agent

    @functools.cached_property
    def i_observed_opponent_by_i_agent(self) -> dict[int, int]:
        if self.coop_config.observation_accuracy == 1:
            return self.i_opponent_by_i_agent
        else:
            whether_resolution = 100
            randomishes_whether = misc.int_to_base(
                int(hashlib.sha1(f'{self.episode_seed}?'.encode()).hexdigest(), 16),
                whether_resolution,
            )
            assert max(randomishes_whether) <= whether_resolution - 1
            assert len(randomishes_whether) >= self.coop_config.n_agents + 1
            randomishes_which = misc.int_to_base(
                int(hashlib.sha1(f'{self.episode_seed}#'.encode()).hexdigest(), 16),
                self.coop_config.n_agents - 1,
            )
            assert max(randomishes_which) <= self.coop_config.n_agents - 2
            assert len(randomishes_which) >= self.coop_config.n_agents + 1
            i_observed_opponent_by_i_agent = self.i_opponent_by_i_agent.copy()
            for i_agent in self.i_agents:
                whether = randomishes_whether[i_agent] / whether_resolution
                if whether >= self.coop_config.observation_accuracy:
                    i_observed_opponent_by_i_agent[i_agent] = i_observed_opponent = \
                                          (i_agent + randomishes_which[i_agent] + 1) % self.n_agents
                    assert i_observed_opponent != i_agent
                    assert i_observed_opponent in self.i_agents

            return i_observed_opponent_by_i_agent

    @functools.cached_property
    def moves_by_i_agent(self) -> tuple[tuple[Move, ...], tuple[Move, ...]]:
        if self.i_turn == 0:
            return ((),) * len(self.agents)
        return tuple(zip(*self.joint_moves, strict=True))

    @functools.cached_property
    def rewards_by_i_agent(self) -> tuple[tuple[RealNumber, ...], tuple[RealNumber, ...]]:
        if self.i_turn == 0:
            return ((),) * len(self.agents)
        return tuple(zip(*self.joint_rewards, strict=True))

    @functools.cached_property
    def reward_by_agent(self) -> dict[Agent, RealNumber]:
        return dict(zip(self.agents, self.joint_rewards[-1], strict=True))

    @functools.cached_property
    def terminated_by_agent(self) -> dict[Agent, bool]:
        return {agent: False for agent in self.agents + (ALL_AGENTS,)}

    @functools.cached_property
    def truncated_by_agent(self) -> dict[Agent, bool]:
        return {agent: (self.i_turn >= self.coop_config.episode_length)
                for agent in self.agents + (ALL_AGENTS,)}

    @staticmethod
    def make_initial(coop_config: CoopConfig) -> State:
        return State(
            coop_config=coop_config,
            i_turn=0,
            joint_moves=(),
            joint_rewards=(),
            i_agent_pairs=tuple(
                map(tuple,
                    map(sorted,
                        more_itertools.chunked(misc.shuffled(range(coop_config.n_agents)), 2)
                    )
                )
            ),
            episode_seed=random.randint(0, sys.maxsize),
        )

    def step(self, actions: Mapping[Agent, np.ndarray]) -> State:

        i_turn = self.i_turn + 1

        joint_move = tuple(Move.from_neural(actions[agent]) for agent in self.agents)
        joint_reward = [None] * len(self.agents)

        for i_left_agent, i_right_agent in self.i_agent_pairs:
            match (joint_move[i_left_agent], joint_move[i_right_agent]):
                case (Move.HAWK, Move.HAWK):
                    joint_reward[i_left_agent] = self.coop_config.anguish_reward
                    joint_reward[i_right_agent] = self.coop_config.anguish_reward
                case (Move.HAWK, Move.DOVE):
                    joint_reward[i_left_agent] = self.coop_config.victory_reward
                    joint_reward[i_right_agent] = self.coop_config.defeat_reward
                case (Move.DOVE, Move.HAWK):
                    joint_reward[i_left_agent] = self.coop_config.defeat_reward
                    joint_reward[i_right_agent] = self.coop_config.victory_reward
                case (Move.DOVE, Move.DOVE):
                    joint_reward[i_left_agent] = self.coop_config.boredom_reward
                    joint_reward[i_right_agent] = self.coop_config.boredom_reward
                case _:
                    raise RuntimeError

        assert None not in joint_reward

        return State(
            coop_config=self.coop_config,
            i_turn=i_turn,
            joint_moves=self.joint_moves + (joint_move,),
            joint_rewards=self.joint_rewards + (tuple(joint_reward),),
            i_agent_pairs=self.i_agent_pairs,
            episode_seed=self.episode_seed,
        )


    @functools.cached_property
    def text(self) -> str:
        if self.i_turn == 0:
            return ','.join(f'{self.coop_config.pad_i_agent(i_left_agent)} '
                            f'{self.coop_config.pad_i_agent(i_right_agent)}'
                            for i_left_agent, i_right_agent in self.i_agent_pairs
            )
        else:
            return ','.join(
                (self.joint_moves[-1][i_left_agent].as_letter +
                 self.joint_moves[-1][i_right_agent].as_letter)
                for i_left_agent, i_right_agent in self.i_agent_pairs
            )
        # if self.i_turn == 0:
            # return '--'
        # else:
            # return f'{self.joint_moves[-1][0].as_letter}{self.joint_moves[-1][1].as_letter}'


    @functools.cached_property
    def zeitgeist(self) -> Zeitgeist:
        if self.i_turn == 0:
            return dict(zip(itertools.combinations(self.i_agents, 2),
                            itertools.repeat(Dialog.make_zero(self.coop_config))))
        dialog_by_i_agent_pair = {}
        for i_agent_a, i_agent_b in self.coop_config.i_agent_pairs_increasing:
            if (i_agent_a, i_agent_b) in self.i_agent_pairs:
                agent_a_rewards = self.rewards_by_i_agent[i_agent_a]
                agent_b_rewards = self.rewards_by_i_agent[i_agent_b]
                dialog_by_i_agent_pair[i_agent_a, i_agent_b] = Dialog(
                    n_agent_a_victories=agent_a_rewards.count(self.coop_config.victory_reward),
                    n_agent_b_victories=agent_b_rewards.count(self.coop_config.victory_reward),
                    n_anguishes=agent_a_rewards.count(self.coop_config.anguish_reward),
                    n_boredoms=agent_a_rewards.count(self.coop_config.boredom_reward),
                    coop_config=self.coop_config,
                )
            else:
                dialog_by_i_agent_pair[i_agent_a, i_agent_b] = Dialog.make_zero(self.coop_config)
        for i_agent_a, i_agent_b in self.coop_config.i_agent_pairs_decreasing:
            dialog_by_i_agent_pair[i_agent_a, i_agent_b] = -dialog_by_i_agent_pair[i_agent_b,
                                                                                   i_agent_a]
        for i_agent_a, i_agent_b in self.coop_config.i_agent_pairs_diagonal:
            assert i_agent_a == i_agent_b
            dialog_by_i_agent_pair[i_agent_a, i_agent_b] = Dialog.make_zero(self.coop_config)


        return Zeitgeist.make_from_dialog_by_i_agent_pair(
            dialog_by_i_agent_pair=dialog_by_i_agent_pair
        )




from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, TypeVar, Annotated
import pathlib
import abc
import math
import statistics
import enum
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import functools
import dataclasses

import chicken_coop # Imported first to filter out warnings.

import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
from frozendict import frozendict
import numpy as np
from scipy import stats
import more_itertools
import networkx as nx
import plotly.graph_objects as go

from chicken_coop import DominanceHierarchy
from chicken_coop.county import nx_tools


def test():
    dominance_hierarchy = DominanceHierarchy('0 => 1 => 2')
    assert dominance_hierarchy == DominanceHierarchy(((0, 1), (0, 2), (1, 2)))
    assert dominance_hierarchy != DominanceHierarchy(((0, 1), (0, 2)))
    assert dominance_hierarchy != DominanceHierarchy(((0, 1), (2, 0), (1, 2)))
    assert str(dominance_hierarchy) == '0 => 1 => 2'
    assert repr(dominance_hierarchy) == "DominanceHierarchy('0 => 1 => 2')"
    assert dominance_hierarchy.is_transitive
    assert dominance_hierarchy.is_complete


def test_distance():
    dh_0 = DominanceHierarchy('0 => 1 => 2 => 3')
    dh_1 = DominanceHierarchy('0 => 3 => 2 => 1')
    dh_2 = DominanceHierarchy('(0 -> 1 -> 2 -> 0) => 3')

    assert (dh_0 - dh_0) == (dh_1 - dh_1) == (dh_2 - dh_2) == 0
    assert (dh_0 - dh_1) == (dh_1 - dh_0) == 3
    assert (dh_0 - dh_2) == (dh_2 - dh_0) == 1
    assert (dh_1 - dh_2) == (dh_2 - dh_1) == 4


def test_isolated():
    dh_0 = DominanceHierarchy('0 => 1 => 2 <3? 4?>')
    dh_1 = DominanceHierarchy('0 => 1 => 2')
    assert dh_0 == dh_0
    assert dh_0.n_agents == 5
    assert dh_1 == dh_1
    assert dh_1.n_agents == 3

    assert dh_0 == DominanceHierarchy(((0, 1), (0, 2), (1, 2)), n_agents=5)
    assert dh_0 != DominanceHierarchy(((0, 1), (0, 2), (1, 2)), n_agents=6)
    assert str(dh_0) == '0 => 1 => 2 <3? 4?>'

def make_di_graph(edges_and_extra_nodes: Iterable[tuple[int, int] | int]):
    edges = []
    extra_nodes = []
    for edge_or_extra_node in edges_and_extra_nodes:
        if isinstance(edge_or_extra_node, tuple):
            edges.append(edge_or_extra_node)
        else:
            assert isinstance(edge_or_extra_node, int)
            extra_nodes.append(edge_or_extra_node)

    di_graph = nx.DiGraph(edges)
    di_graph.add_nodes_from(extra_nodes)
    return di_graph


def test_str():
    def test_case(di_graph: nx.DiGraph, concise_text: str) -> None:
        assert nx_tools.oriented_graph_to_concise_text(di_graph) == concise_text
        assert nx.utils.graphs_equal(di_graph,
                                     nx_tools.concise_text_to_oriented_graph(concise_text))

    test_case(
        nx.DiGraph([(0, 1), (0, 2), (0, 3), (0, 4), (0, 5),
                    (4, 1), (4, 2), (4, 3), (4, 5),
                    (1, 2), (1, 3), (1, 5),
                    (2, 3), (2, 5),
                    (5, 3)]),
        '0 => 4 => 1 => 2 => 5 => 3',
    )
    test_case(
        nx.DiGraph([
            (1, 2), (2, 3), (3, 1),
            (1, 4), (1, 0), (1, 5),
            (2, 4), (2, 0), (2, 5),
            (3, 4), (3, 0), (3, 5),
            (4, 0), (4, 5), (0, 5)
        ]),
        '(1 -> 2 -> 3 -> 1) => 4 => 0 => 5',
    )
    test_case(
        nx.DiGraph([(1, 2), (2, 3), (3, 1),
                    (1, 0), (1, 4), (1, 5),
                    (2, 0), (2, 4), (2, 5),
                    (3, 0), (3, 4), (3, 5),
                    (0, 4), (4, 5), (5, 0)]),
        '(1 -> 2 -> 3 -> 1) => (0 -> 4 -> 5 -> 0)',
    )
    test_case(
        nx.DiGraph([
            (0, 1), (0, 2), (0, 3), (0, 4), (0, 5),
            (1, 3), (1, 4),
            (2, 1),
            (3, 2),
            (4, 2), (4, 3),
            (1, 5), (2, 5), (3, 5), (4, 5),
        ]),
        '0 => (1 -> 3 -> 2 -> 1 -> 4 -> 2 | 4 -> 3) => 5',
    )
    test_case(
        make_di_graph([
            0,
            (1, 3),
            (2, 1),
            (3, 2),
            (4, 1), (4, 2), (4, 3),
        ]),
        '4 => (1 -> 3 -> 2 -> 1) <0?>',
    )
    test_case(
        make_di_graph([
            (0, 1),
            (2, 4),
            3,
            5,
            6,
        ]),
        '(0, 1), (2, 4) <3? 5? 6?>',
    )


def test_distance():
    assert DominanceHierarchy('0 => 1 => 2') - DominanceHierarchy('2 => 1 => 0') == \
           DominanceHierarchy('2 => 1 => 0') - DominanceHierarchy('0 => 1 => 2') == \
           DominanceHierarchy('2 => 1 => 0').get_distance(DominanceHierarchy('0 => 1 => 2')) == \
           DominanceHierarchy('2 => 1 => 0').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 1, 2)) == \
           DominanceHierarchy('2 => 1 => 0').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 1)) == \
           DominanceHierarchy('2 => 1 => 0').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 2)) == \
           DominanceHierarchy('2 => 1 => 0').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(1, 2)) == 1

    assert DominanceHierarchy('0 => 1 => 2') - DominanceHierarchy('0 => 1 => 2') == \
           DominanceHierarchy('0 => 1 => 2') - DominanceHierarchy('0 => 1 => 2') == \
           DominanceHierarchy('0 => 1 => 2').get_distance(DominanceHierarchy('0 => 1 => 2')) == \
           DominanceHierarchy('0 => 1 => 2').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 1, 2)) == \
           DominanceHierarchy('0 => 1 => 2').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 1)) == \
           DominanceHierarchy('0 => 1 => 2').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(0, 2)) == \
           DominanceHierarchy('0 => 1 => 2').get_distance(DominanceHierarchy('0 => 1 => 2'),
                                                          i_agents=(1, 2)) == 0

    assert DominanceHierarchy('0 => 1 => 2 => 3') - DominanceHierarchy('3 => 2 => 1 => 0') == \
           DominanceHierarchy('3 => 2 => 1 => 0') - DominanceHierarchy('0 => 1 => 2 => 3') == 1

    assert DominanceHierarchy('0 => 1 => 2 => 3') - DominanceHierarchy('0 => 1 => 3 => 2') == 1 / 6

    assert DominanceHierarchy('0 => 1 => 2 => 3').get_distance(
                                    DominanceHierarchy('0 => 1 => 3 => 2'), i_agents=(0, 1, 2)) == 0
    assert DominanceHierarchy('0 => 1 => 2 => 3').get_distance(
                                DominanceHierarchy('0 => 1 => 3 => 2'), i_agents=(1, 2, 3)) == 1 / 3
    assert DominanceHierarchy('0 => 1 => 2 => 3').get_distance(
                                DominanceHierarchy('0 => 1 => 3 => 2'), i_agents=(2, 3)) == 1




import sys
from typing import TextIO, Iterable
import logging
import re
import warnings

class FiltrosStream:
    def __init__(self, original_stream: TextIO, patterns: Iterable[str] = ()) -> None:
        self.original_stream = original_stream
        self.patterns = tuple(patterns)
        self.combined_pattern = re.compile('|'.join(f'(?:{pattern})' for pattern in patterns))
        self.last_printed = True

    def write(self, message: str) -> None:
        # print(f'XXX: {repr(message)}')
        if message.isspace() and not self.last_printed:
            return
        if self.combined_pattern.search(message):
            self.last_printed = False
            return
        else:
            self.last_printed = True
            self.original_stream.write(message)

    def flush(self) -> None:
        self.original_stream.flush()

    def fileno(self) -> int:
        return self.original_stream.fileno()

    def close(self) -> None:
        pass


def filtros(patterns) -> None:
    sys.stderr = FiltrosStream(sys.stderr, patterns)


def activate() -> None:
    warnings.filterwarnings('ignore', '`np.bool8` is a deprecated alias for `np.bool_`',
                            DeprecationWarning)
    filtros(
        (
            r'The distutils package is deprecated',
            r'import distutils',

            r"Setting 'object_store_memory' for actors",

            r'DeprecationWarning: non-integer arguments to randrange\(\)',
            r'self.episode_id: int = random\.randrange\(2e9\)',

            r'Install gputil for GPU system monitoring',

            r'ray/rllib.* DeprecationWarning: `np\.bool` is a deprecated alias for the',
            r'Deprecated in NumPy 1\.20; for more details and guidance',
            r'if not isinstance\(done_, \(bool, np\.bool, np\.bool_\)\):',

            r'/ray/dashboard.* DeprecationWarning: There is no current event loop',
            r'aiogrpc\.init_grpc_aio\(\)',
            r'loop = asyncio\.get_event_loop\(\)',

            r'(?:observation|action)_space_(?:contains|sample)\(\) has not been implemented',

            r'DeprecationWarning: `_get_slice_indices` has been deprecated',

            r'Current log_level is .{1,20}For more information,',

            r'[0-9]{2},[0-9]{3}\s+INFO ',

            r'Function `rng.randint\(low, \[high, size, dtype\]\)` is marked as deprecated',

            r'past/.*DeprecationWarning: the imp module',
            r'from imp import reload',

            r'DeprecationWarning: `tune.report` and `tune.checkpoint_dir` APIs are deprecated',

            r'train_batch_size.*cannot be achieved',

            r'Function checkpointing is disabled\. This may result',

            (r'DeprecationWarning: `(DirectStepOptimizer|build_policy_class|build_tf_policy|'
             r'ValueNetworkMixin|LearningRateSchedule|EntropyCoeffSchedule|KLCoeffMixin|'
             r'remote_workers\(\))` '
             r'has been deprecated'),

            r'Setting `exploration_config={}` because you set `_enable_rl_module_api=True`',

            r'ray/rllib/algorithms/algorithm.py:484: RayDeprecationWarning: This API is deprecated',
            r'`UnifiedLogger` will be removed',

            r'ray/tune/logger/unified.py:53: RayDeprecationWarning: This API is deprecated',
            r'The `[a-zA-Z0-9]{1,40}Logger`? interface is deprecated',
            r'self._loggers.append(cls(self.config, self.logdir, self.trial))',

        )
    )




from __future__ import annotations

import random
import re
import tempfile
import time
import math
import os
import logging
import pathlib
import sys
import itertools
import warnings
import functools
from typing import Optional
import pathlib
import webbrowser

import click
import more_itertools
import pandas as pd
import numpy as np
import scipy.signal
import networkx as nx

import chicken_coop

from ..command_group import cli
from chicken_coop.county import misc
from chicken_coop.county import trekking
from chicken_coop.county.typing import RealNumber
from .. import utils
from ..dominance_hierarching import DominanceHierarchy
from . import constants


@cli.group()
@click.option('-p', '--path', 'trek_path_expression', type=str, required=False)
@click.pass_context
def plot(context: click.Context, *, trek_path_expression: Optional[str]) -> None:
    context.ensure_object(dict)
    context.obj['trek'] = trekking.Trek.get(trek_path_expression, allow_multi_trek=False)


@plot.command()
@click.option('-s', '--smoothing', type=int, required=False)
@click.pass_context
def plain(context: click.Context, *, smoothing: Optional[int]) -> None:
    import plotly.graph_objects as go
    trek = context.obj['trek']
    (mini_trek,) = trek.mini_treks

    print(f'Making a plot for {mini_trek} ...', file=sys.stderr)
    df = mini_trek.rollout_dataframe
    axis_template = {'title_font': {'size': 23,}, 'tickfont': {'size': 20,}}
    interesting_column_names = tuple(name for name in df.columns if 'agreeability' in name)
    if smoothing is None:
        wrapper = lambda column: column
    else:
        wrapper = lambda column: scipy.signal.savgol_filter(
            np.array(column),
            smoothing, # window size used for filtering
            3, # order of fitted polynomial
        )
    figure = go.Figure(
        data=tuple(
            go.Scatter(x=df.index, y=wrapper(df[interesting_column_name]),
                       name=interesting_column_name)
            for interesting_column_name in interesting_column_names
        ),
        layout=go.Layout(
            title_font_size=28,
            xaxis={**axis_template, 'title': 'Generation'},
            yaxis={**axis_template, 'title': ''},
            legend={#'xanchor': 'right', 'x': 1, 'yanchor': 'bottom',
                    #'y': 0,
                    'font': {'size': 20}}
        )
    )
    figure.show()


@plot.command()
@click.option('-o', '--output', is_flag=True)
@click.pass_context
def dh(context, *, output: bool) -> None:
    import plotly.graph_objects as go
    trek = context.obj['trek']
    (mini_trek,) = trek.mini_treks

    print(f'Making a dominance hierarchy plot for {mini_trek} ...', file=sys.stderr)
    df = mini_trek.rollout_dataframe
    # dominance_hierarchy = DominanceHierarchy(df.iloc[-1]['aggregated_dominance_hierarchy'])
    dominance_hierarchy = DominanceHierarchy(
        # '(1 -> 2 -> 3 -> 1) => 4 => 0 => (6 -> 7 -> 8 -> 9 -> 6 -> 8 | 9 -> 7) => 5'
        # '(0 -> 1 -> 2 -> 3 -> 4 -> 0 | 2 -> 0 -> 3 -> 1 -> 4 -> 2) => 5'
        # '0 => 1 => 2 => 3 => 4 => 5 => 6'
        # '0 => 1 => (2 -> 3 -> 4 -> 2) => 5 => 6'
        # '0 => (1 -> 2 -> 3 -> 4 -> 1 | 1 -> 3 | 2 -> 4) => 5 => 6'
        '0 => (1 -> 2 -> 3 -> 1) => (4 -> 5 -> 6 -> 4)'
    )
    if output:
        click.echo(dominance_hierarchy.condensed_plot_svg)
    else:
        with tempfile.NamedTemporaryFile('w', suffix='.html') as fp:
            indent = ' ' * 8
            fp.write(
                constants.HTML_TEMPLATE.format(
                    indent + dominance_hierarchy.condensed_plot_svg.replace('\n', '\n' + indent)
                )
            )
            fp.flush()
            webbrowser.open_new_tab(fp.name)
            time.sleep(3)



# @plot.command()
# # @click.option('-p', '--path', 'path_expression', type=str, required=False)
# @click.option('-c', '--color', type=str, default='blue')
# @click.argument('data', type=str)
# @click.pass_context
# def plot_poster_rank_linearity(*, data: str, color: str) -> None:
    # # 0.7910447761194029,0.4925373134328358,0.19402985074626866,0.08955223880597014,0.014925373134328358,0.0,0.0,0.029850746268656716,0.029850746268656716,0.0,0.13432835820895522,0.40298507462686567 --color blue
    # # 1.0,0.8,0.6,0.26666666666666666,0.2,0.2,0.13333333333333333,0.26666666666666666,0.2,0.26666666666666666,0.3333333333333333,0.5333333333333333 --color red
    # import plotly.graph_objects as go
    # print(f'Making a poster rank linearity plot...', file=sys.stderr)
    # rank_legibilities = tuple(map(float, data.split(',')))
    # for rank_linearity in rank_legibilities:
        # assert 0 <= rank_linearity <= 1
    # df = pd.DataFrame(rank_legibilities, columns=('rank_linearity',))
    # axis_template = {'title_font': {'size': 350,}, 'tickfont': {'size': 175,}}
    # figure = go.Figure(
        # data=go.Scatter(
            # x=df.index, y=df['rank_linearity'], name='Rank linearity',
            # mode='lines+markers',
            # line=dict(width=25),
            # marker=dict(
                # color=color,
                # size=125,
            # )
        # ),
        # layout=go.Layout(
            # title_font_size=500,
            # xaxis={**axis_template,
                   # 'title': dict(
                       # text='Social rank',
                       # standoff=150,
                   # ),
                   # 'tickvals': tuple(range(len(rank_legibilities) + 1)),},
            # yaxis={**axis_template,
                   # 'title': dict(
                       # text='Rank linearity',
                       # standoff=150,
                   # ),
                   # 'range': (-0.05, 1.05),
                   # 'tickvals': np.arange(0, 1.1, 0.1),},
            # autosize=False,
            # width=5_000,
            # height=5_000,
            # # legend={#'xanchor': 'right', 'x': 1, 'yanchor': 'bottom',
                    # #'y': 0,
                    # # 'font': {'size': 20}}
        # )
    # )
    # figure.show()



# @plot.command()
# @click.argument('csv_path_string', type=click.Path(exists=True, file_okay=True, dir_okay=False))
# @click.pass_context
# def plot_poster_aggressiveness_animals(*, csv_path_string: str) -> None:
    # import plotly.graph_objects as go
    # print(f'Making a poster animal aggressiveness plot...', file=sys.stderr)
    # df = pd.read_csv(csv_path_string).drop(columns='time_step')
    # axis_template = {'title_font': {'size': 350,}, 'tickfont': {'size': 175,}}
    # figure = go.Figure(
        # data=[
            # go.Scatter(
                # x=df.index, y=df[f'{i + 1}_aggressiveness'], name=f'Agent {i}',
                # mode='lines',
                # line=dict(width=18),
            # ) for i in range(len(df.columns))
        # ],
        # layout=go.Layout(
            # title_font_size=500,
            # xaxis={**axis_template,
                   # 'title': dict(
                       # text='Time',
                       # standoff=150,
                   # )},
            # yaxis={**axis_template,
                   # 'title': dict(
                       # text='Aggressiveness',
                       # standoff=150,
                   # ),
                   # 'range': (-0.05, 1.05),
                   # 'tickvals': np.arange(0, 1.1, 0.1),},
            # autosize=False,
            # width=5_000,
            # height=5_000,
            # legend={'xanchor': 'right',
                    # 'x': 1,
                    # 'yanchor': 'bottom',
                    # 'y': 0.7,
                    # 'entrywidth': 100,
                    # 'entrywidthmode': 'pixels',
                    # 'itemwidth': 100,
                    # 'borderwidth': 0,
                    # 'font': {'size': 100}}
        # )
    # )
    # figure.show()



# @plot.command()
# @click.argument('jsonl_path_string', type=click.Path(exists=True, file_okay=True, dir_okay=False))
# def plot_poster_aggressiveness_ai(*, jsonl_path_string: str) -> None:
    # import plotly.graph_objects as go
    # print(f'Making a poster AI aggressiveness plot...', file=sys.stderr)
    # df = misc.jsonl_to_dataframe(jsonl_path_string, dot_notation=True)
    # df = df[
        # ['generation'] +
        # sorted(filter(re.compile(r'^agent\.[0-9]+\.aggressiveness$').fullmatch,
                      # df.columns))
    # ]
    # axis_template = {'title_font': {'size': 350,}, 'tickfont': {'size': 175,}}
    # figure = go.Figure(
        # data=[
            # go.Scatter(
                # x=df.index, y=df[f'agent.{i}.aggressiveness'], name=f'Agent {i}',
                # mode='lines',
                # line=dict(width=18),
            # ) for i in range(len(df.columns) - 1)
        # ],
        # layout=go.Layout(
            # title_font_size=500,
            # xaxis={**axis_template,
                   # 'title': dict(
                       # text='Time',
                       # standoff=150,
                   # )},
            # yaxis={**axis_template,
                   # 'title': dict(
                       # text='Aggressiveness',
                       # standoff=150,
                   # ),
                   # 'range': (-0.05, 1.05),
                   # 'tickvals': np.arange(0, 1.1, 0.1),},
            # autosize=False,
            # width=5_000,
            # height=5_000,
            # legend={'xanchor': 'right',
                    # 'x': 1,
                    # 'yanchor': 'bottom',
                    # 'y': 0.7,
                    # 'entrywidth': 100,
                    # 'entrywidthmode': 'pixels',
                    # 'itemwidth': 100,
                    # 'borderwidth': 0,
                    # 'font': {'size': 100}}
        # )
    # )
    # figure.show()






import sys
import os
import pathlib
import importlib

import pytest


def __bootstrap():
    '''
    Add needed packages in repo to path if we can't find them.

    This adds `chicken_coop`'s root folder to `sys.path` if it can't
    currently be imported.
    '''
    if not importlib.util.find_spec('chicken_coop'):
        chicken_coop_candidate_path = pathlib(__file__).parent.parent.absolute()
        sys.path.append(chicken_coop_candidate_path)


__bootstrap()


def invoke_tests():
    os.chdir(os.path.dirname(__file__))
    pytest.main()


from __future__ import annotations

import abc
import random
import math
import os
import logging
import pathlib
import sys
import itertools
import tempfile
import warnings
import functools
from typing import Optional, Iterable
import pathlib

import click
import more_itertools
import pandas as pd
import numpy as np
import scipy.signal
import networkx as nx
import plotly.graph_objects as go
import svgwrite

import chicken_coop

from ..command_group import cli
from chicken_coop.county import nx_tools
from chicken_coop.county import misc
from chicken_coop.county.typing import RealNumber
from .. import utils
from ..dominance_hierarching import DominanceHierarchy


DOUBLE_ARROW_LENGTH = 10
DOUBLE_ARROW_GAP = 1.8
DOUBLE_ARROW_HEAD_DIAGONAL_LENGTH = 4.5
DOUBLE_ARROW_SHAFT_STROKE_WIDTH = 0.7
DOUBLE_ARROW_HEAD_STROKE_WIDTH = 1
SMALL_ARROW_GAP = 8
SMALL_ARROW_STROKE_WIDTH = 0.7
SMALL_ARROW_HEAD_DIAGONAL_LENGTH = 2.5
PADDING = 2.2
NODE_RADIUS = 5

RED = '#c66'

FONT_X_OFFSET = 0
FONT_Y_OFFSET = 1
FONT_SIZE = 8
FONT_FAMILY = '"Noto Mono", Consolas, "Ubuntu Mono", monospace'



def make_condensed_plot_svg(dominance_hierarchies: Iterable[DominanceHierarchy] | DominanceHierarchy
                            ) -> str:
    dominance_hierarchies = misc.to_tuple(dominance_hierarchies, item_type=DominanceHierarchy)
    return StackedCondensedDominanceHierarchyShapino(dominance_hierarchies).svg_text


class CuteShapino(misc.Shapino, abc.ABC):
    def __init__(self, *args, x: RealNumber = 0, y: RealNumber = 0, **kwargs) -> None:
        super().__init__(self.make_cute_shapes(*args, **kwargs), x=x, y=y)

    @abc.abstractmethod
    def make_cute_shapes(self, *args, **kwargs
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        raise NotImplementedError

class RowShapino(CuteShapino):

    padding = 0

    def make_cute_shapes(self, shapinos: Iterable[misc.Shapino],
                         debug: bool = False,
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        shapinos = tuple(shapinos)
        new_shapinos = [
            shapino.duplicate_with_anchor_on_middle_left().duplicate_with_offset(offset, 0)
            for shapino, offset in zip(
                shapinos,
                misc.zero_based_accumulate((shapino.bounding_box[2] + self.padding
                                            for shapino in shapinos),
                                           drop_last=True),
                strict=True
            )
        ]
        if debug:
            for new_shapino in new_shapinos:
                new_shapino.shapes.append(
                    svgwrite.shapes.Rect(
                        new_shapino.bounding_box[:2],
                        new_shapino.bounding_box[2:],
                        stroke=random.choice(('red', 'orange', 'purple', 'yellow', 'green')),
                        stroke_dasharray=2,
                        stroke_width=0.2,
                        fill='blue',
                        fill_opacity=0.1,
                        stroke_opacity=0.2,
                    )
                )
        return new_shapinos


class ColumnShapino(CuteShapino):

    padding = 0

    def make_cute_shapes(self, shapinos: Iterable[misc.Shapino],
                         debug: bool = False,
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        shapinos = tuple(shapinos)
        new_shapinos = [
            shapino.duplicate_with_anchor_on_top_center().duplicate_with_offset(0, offset)
            for shapino, offset in zip(
                shapinos,
                misc.zero_based_accumulate((shapino.bounding_box[3] + self.padding
                                            for shapino in shapinos),
                                           drop_last=True),
                strict=True
            )
        ]
        if debug:
            for new_shapino in new_shapinos:
                new_shapino.shapes.append(
                    svgwrite.shapes.Rect(
                        new_shapino.bounding_box[:2],
                        new_shapino.bounding_box[2:],
                        stroke=random.choice(('red', 'orange', 'purple', 'yellow', 'green')),
                        stroke_dasharray=2,
                        stroke_width=0.2,
                        fill='blue',
                        fill_opacity=0.1,
                        stroke_opacity=0.2,
                    )
                )
        return new_shapinos



class StackedCondensedDominanceHierarchyShapino(ColumnShapino):

    padding = 8

    def make_cute_shapes(self, dominance_hierarchies: Iterable[DominanceHierarchy]
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        return ColumnShapino.make_cute_shapes(
            self,
            [CondensedDominanceHierarchyShapino(dominance_hierarchy)
             for dominance_hierarchy in dominance_hierarchies]
        )


class CondensedDominanceHierarchyShapino(RowShapino):
    def make_cute_shapes(self, dominance_hierarchy: DominanceHierarchy
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        shapinos = []
        for i, condensed_node in enumerate(nx.topological_sort(
                                                           dominance_hierarchy.condensed_di_graph)):
            nodes = dominance_hierarchy.condensed_di_graph.nodes.data()[condensed_node]['members']
            shapinos.append(SccShapino(nx_tools.get_sub_di_graph(dominance_hierarchy.di_graph,
                                                                 nodes)))
            if i < len(dominance_hierarchy.condensed_di_graph) - 1:
                shapinos.append(DoubleArrowShapino())
        return RowShapino.make_cute_shapes(self, shapinos)


class NodeShapino(CuteShapino):

    padding = PADDING

    def make_cute_shapes(self, text: str) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        assert len(text) == 1
        self.text = text
        return (
            svgwrite.shapes.Circle(r=NODE_RADIUS, fill='#cce', stroke='#666', stroke_width=0.5),
            svgwrite.text.Text(text=text, font_size=FONT_SIZE, font_family=FONT_FAMILY,
                               text_anchor='middle', dominant_baseline='middle',
                               fill='#444',
                               insert=(FONT_X_OFFSET, FONT_Y_OFFSET),
                               font_weight='bold'),
        )



class SccShapino(CuteShapino):

    padding = PADDING

    def make_cute_shapes(self, scc: nx.DiGraph
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        self.scc = scc
        nodes = sorted(scc)
        n_nodes = len(scc)
        self.is_single = (n_nodes == 1)
        radius = _get_scc_radius(scc)
        node_shapinos = [
            NodeShapino(
                x=radius * math.cos(math.tau * i / n_nodes),
                y=radius * math.sin(math.tau * i / n_nodes),
                text=str(node),
            )
            for i, node in enumerate(nodes)
        ]

        small_arrow_shapinos = []
        for dominant_node, subordinate_node in scc.edges:
            i_dominant_node = nodes.index(dominant_node)
            i_subordinate_node = nodes.index(subordinate_node)
            gapless_source_x = radius * math.cos(math.tau * i_dominant_node / n_nodes)
            gapless_source_y = radius * math.sin(math.tau * i_dominant_node / n_nodes)
            gapless_destination_x = radius * math.cos(math.tau * i_subordinate_node / n_nodes)
            gapless_destination_y = radius * math.sin(math.tau * i_subordinate_node / n_nodes)
            gapless_delta_x = gapless_destination_x - gapless_source_x
            gapless_delta_y = gapless_destination_y - gapless_source_y
            gapless_length = (gapless_delta_x ** 2 + gapless_delta_y ** 2) ** 0.5
            gap_x = gapless_delta_x * SMALL_ARROW_GAP / gapless_length
            gap_y = gapless_delta_y * SMALL_ARROW_GAP / gapless_length

            small_arrow_shapinos.append(
                SmallArrowShapino(
                    source_x=gapless_source_x + gap_x,
                    source_y=gapless_source_y + gap_y,
                    destination_x=gapless_destination_x - gap_x,
                    destination_y=gapless_destination_y - gap_y,
                )
            )

        shapinos = node_shapinos + small_arrow_shapinos
        if not self.is_single:
            big_circle = svgwrite.shapes.Circle(
                r=radius + NODE_RADIUS + 3,
                fill=RED,
                opacity=0.18,
            )
            shapinos.insert(0, big_circle)
        return shapinos




class SingleArrowShapino(CuteShapino):

    def make_cute_shapes(self, start: tuple[RealNumber, RealNumber],
                         end: tuple[RealNumber, RealNumber]
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        raise NotImplementedError
        stroke_kwargs = {'stroke': RED}
        shaft_lines = [
            svgwrite.shapes.Line(
                start=(0, sign * DOUBLE_ARROW_GAP / 2),
                end=(DOUBLE_ARROW_LENGTH - DOUBLE_ARROW_GAP / 2, sign * DOUBLE_ARROW_GAP / 2),
                stroke_width=DOUBLE_ARROW_SHAFT_STROKE_WIDTH,
                **stroke_kwargs,
            )
            for sign in (-1, 1)
        ]
        foo = DOUBLE_ARROW_HEAD_DIAGONAL_LENGTH * pow(2, -0.5)
        head_lines = [
            svgwrite.shapes.Line(
                start=(DOUBLE_ARROW_LENGTH, 0),
                end=(DOUBLE_ARROW_LENGTH - foo , sign * foo),
                stroke_linecap='round',
                stroke_width=DOUBLE_ARROW_HEAD_STROKE_WIDTH,
                **stroke_kwargs,
            )
            for sign in (-1, 1)
        ]
        return (*shaft_lines, *head_lines)


class DoubleArrowShapino(CuteShapino):

    padding = PADDING

    def make_cute_shapes(self) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        stroke_kwargs = {'stroke': RED}
        shaft_lines = [
            svgwrite.shapes.Line(
                start=(0, sign * DOUBLE_ARROW_GAP / 2),
                end=(DOUBLE_ARROW_LENGTH - DOUBLE_ARROW_GAP / 2, sign * DOUBLE_ARROW_GAP / 2),
                stroke_width=DOUBLE_ARROW_SHAFT_STROKE_WIDTH,
                **stroke_kwargs,
            )
            for sign in (-1, 1)
        ]
        foo = DOUBLE_ARROW_HEAD_DIAGONAL_LENGTH * pow(2, -0.5)
        head_lines = [
            svgwrite.shapes.Line(
                start=(DOUBLE_ARROW_LENGTH, 0),
                end=(DOUBLE_ARROW_LENGTH - foo , sign * foo),
                stroke_linecap='round',
                stroke_width=DOUBLE_ARROW_HEAD_STROKE_WIDTH,
                **stroke_kwargs,
            )
            for sign in (-1, 1)
        ]
        return (*shaft_lines, *head_lines)


class SmallArrowShapino(CuteShapino):

    padding = PADDING

    def make_cute_shapes(self, source_x: RealNumber, source_y: RealNumber,
                         destination_x: RealNumber, destination_y: RealNumber
                         ) -> Iterable[svgwrite.base.BaseElement | misc.Shapino]:
        stroke_kwargs = {'stroke': RED,
                         'stroke_width': SMALL_ARROW_STROKE_WIDTH}
        shaft_line = svgwrite.shapes.Line(
                start=(source_x, source_y),
                end=(destination_x, destination_y),
                **stroke_kwargs
            )

        delta_x = source_x - destination_x
        delta_y = source_y - destination_y
        length = (delta_x ** 2 + delta_y ** 2) ** 0.5
        foo = SMALL_ARROW_HEAD_DIAGONAL_LENGTH * pow(2, -0.5)
        delta_fluff_x = foo * delta_x / length
        delta_fluff_y = foo * delta_y / length
        angle = math.atan2(delta_fluff_y, delta_fluff_x) + math.tau / 4
        delta_zuff_x = foo * math.cos(angle)
        delta_zuff_y = foo * math.sin(angle)

        head_lines = [
            svgwrite.shapes.Line(
                start=(destination_x, destination_y),
                end=(destination_x + delta_fluff_x + sign * delta_zuff_x,
                     destination_y + delta_fluff_y + sign * delta_zuff_y,),
                stroke_linecap='round',
                **stroke_kwargs,
            )
            for sign in (-1, 1)
        ]

        return [shaft_line] + head_lines



def _get_scc_radius(scc: nx.DiGraph) -> RealNumber:
    return ((0 if (len(scc) == 1) else 14 + 5 * (len(scc) - 3)))




from __future__ import annotations

import time
import sys
import multiprocessing
from typing import ContextManager, Optional, Type
from types import TracebackType as Traceback

class WingDebuggerPause(ContextManager):
    def __enter__(self) -> WingDebuggerPause:
        try:
            wing_debugger = sys._wing_debugger # pytype: disable=module-attr
        except AttributeError:
            pass
        else:
            wing_debugger.SuspendDebug()
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException],
                 traceback: Optional[Traceback]) -> Optional[bool]:
        try:
            wing_debugger = sys._wing_debugger # pytype: disable=module-attr
        except AttributeError:
            pass
        else:
            wing_debugger.ResumeDebug()



show_import_time = ((multiprocessing.current_process().name == 'MainProcess') and
                    sys.argv[0].endswith('chicken_coop/__main__.py') or
                    sys.argv[0].endswith('chicken_coop') or
                    sys.argv[0].endswith('_wing_run.py'))
start_time = time.monotonic()
if show_import_time:
    print('Starting Chicken Coop, importing heavy modules...')

with WingDebuggerPause():
    from . import county # Importing this before ray because of filtros.

if show_import_time:
    print(f'Done importing heavy modules in {time.monotonic() - start_time:.1f}s.')



from __future__ import annotations

import copy
import functools
import re
import yaml
import tempfile
import contextlib
import pathlib
import operator
import math
import os
import datetime as datetime_module
import lzma
import random
import json
import shutil

import pandas as pd
import sys
import subprocess
import itertools
import operator
import collections.abc
import contextlib
import pathlib
import pickle
import more_itertools
from typing import (Any, Mapping, Optional, BinaryIO, Iterable, TypeVar, TextIO, Iterator, Generator,
                    Hashable, Callable)
import io

import numpy as np
import ray.rllib
import svgwrite

from .constants import CHICKEN_COOP_HOME
from chicken_coop.county import csv_tools
from ..dominance_hierarching import (DominanceHierarchy,
                                     AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy,
                                     AttemptedCompleteOperationOnIncompleteDominanceHierarchy)
from .typing import Agent, Action, RealNumber
from chicken_coop import county
from . import filtros
from . import misc


class BaseTrek:
    folder: pathlib.Path

    @property
    def meta_yaml_path(self) -> pathlib.Path:
        return self.folder / 'meta.yaml'

    @property
    def meta(self) -> dict:
        with self.meta_yaml_path.open('r') as file:
            return yaml.safe_load(file)

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(self.folder)})'

    def __str__(self) -> str:
        return str(self.folder)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BaseTrek):
            return False
        return (type(self) is type(other)) and (self.folder == other.folder)

    def __hash__(self) -> int:
        return hash((type(self), self.folder))


class Trek(BaseTrek):

    @staticmethod
    def create(*, extra_meta: Optional[dict[str, Any]] = None) -> Trek:
        folder = CHICKEN_COOP_HOME / (datetime_module.datetime.now().isoformat()
                                      .replace(':', '-').replace('.', '-').replace('T', '-'))
        folder.mkdir(parents=True, exist_ok=False)
        with (folder / 'meta.yaml').open('w') as yaml_file:
            yaml.dump(
                {
                    'argv': sys.argv,
                } | extra_meta,
                yaml_file,
            )
        return Trek(folder)

    def __init__(self, folder: pathlib.Path | str | os.PathLike) -> None:
        self.folder = pathlib.Path(folder).resolve()
        if not self.folder.exists():
            raise FileNotFoundError(self.folder)
        if not self.folder.is_dir():
            raise NotADirectoryError(self.folder)


    @property
    def moniker(self) -> Optional[str]:
        return self.meta.get('moniker', None)


    def __enter__(self) -> pathlib.Path:
        self.original_stdout = sys.stdout
        sys.stdout = misc.TeeStream(self.original_stdout, self.folder / 'stdout')

        self.original_stderr = sys.stderr
        sys.stderr = misc.TeeStream(self.original_stderr, self.folder / 'stderr')

        print(f'Writing to folder {self.folder}')

        return self

    def __exit__(self, exception_type: Optional[Type[BaseException]],
                 exception: Optional[BaseException],
                 traceback: Optional[Any]) -> None:
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr


    @property
    def mini_treks(self) -> tuple[MiniTrek, ...]:
        return tuple(MiniTrek(path.parent) for path in sorted(self.folder.rglob('rollout.jsonl')))

    def get_mini_treks_grouped_by(self, key_function: Callable[MiniTrek, Any]
                                  ) -> dict[Any, tuple[MiniTrek, ...]]:
        result = collections.defaultdict(list)
        for mini_trek in self.mini_treks:
            value = key_function(mini_trek)
            result[value].append(mini_trek)
        return {value: tuple(mini_treks) for value, mini_treks in result.items()}


    @property
    def is_multi_trek(self) -> bool:
        return len(self.mini_treks) >= 2

    @property
    def is_single_trek(self) -> bool:
        return len(self.mini_treks) == 1

    @property
    def is_empty_trek(self) -> bool:
        return len(self.mini_treks) == 0


    @staticmethod
    def get(trek_path_expression: str | pathlib.Path | None,
            *, allow_multi_trek: bool = True, allow_single_trek: bool = True,
            allow_empty_trek: bool = False, moniker: Optional[str] = None) -> Trek:
        '''
        Get a trek, usually for running analysis or producing plots.

        A trek is a folder in `~/.chicken-coop/` which represents a single run of the `chicken_coop`
        package.

        When the user runs a command, e.g. a plot command, we allow the user to be able to specify
        which trek they are plotting, but for convenience we also allow not specifying a trek, in
        which case we choose the newest trek.
        '''
        trek_allowed = lambda trek: (
            ((allow_multi_trek and trek.is_multi_trek) or
             (allow_single_trek and trek.is_single_trek) or
             (allow_empty_trek and trek.is_empty_trek)) and
            (moniker is None or trek.moniker == moniker)
        )

        if trek_path_expression is None:
            for folder in sorted(CHICKEN_COOP_HOME.iterdir(), reverse=True):
                trek = Trek(folder)
                if trek_allowed(trek):
                    return trek
            else:
                raise FileNotFoundError
        else:
            trek = Trek(trek_path_expression)
            if trek_allowed(trek):
                return trek
            else:
                raise Exception


    def get_sample_mini_treks(self, *, forgive_lack_of_samples: bool = False
                              ) -> tuple[MiniTrek, ...]:
        n_agents = 6
        two_linear_mini_treks = []
        middle_non_linear_mini_trek = None
        non_middle_non_linear_mini_trek = None

        for mini_trek in self.mini_treks:
            dominance_hierarchy: DominanceHierarchy = mini_trek.dominance_hierarchy
            assert dominance_hierarchy.n_agents == n_agents

            if not dominance_hierarchy.is_complete:
                continue
            elif dominance_hierarchy.is_transitive:
                if len(two_linear_mini_treks) <= 1:
                    two_linear_mini_treks.append(mini_trek)
            elif (middle_non_linear_mini_trek is None) and (
                                       {0, n_agents - 1} <= set(dominance_hierarchy.linear_ranks)):
                middle_non_linear_mini_trek = mini_trek
            elif (non_middle_non_linear_mini_trek is None) and \
                                     ({0, n_agents - 1} & set(dominance_hierarchy.non_linear_ranks)):
                non_middle_non_linear_mini_trek = mini_trek

            sample_mini_treks = tuple(two_linear_mini_treks) + (middle_non_linear_mini_trek,
                                                                non_middle_non_linear_mini_trek)
            if len(sample_mini_treks) == 4 and (None not in sample_mini_treks):
                return sample_mini_treks

        if forgive_lack_of_samples and two_linear_mini_treks:
            return (tuple(filter(None, sample_mini_treks)) * 4)[:4]
        else:
            raise Exception




class MiniTrek(BaseTrek):
    def __init__(self, folder: pathlib.Path) -> None:
        self.folder = folder
        self.rollout_jsonl_path = self.folder / 'rollout.jsonl'
        self.progress_csv_path = self.folder / 'progress.csv'


    @functools.cached_property
    def dominance_hierarchy(self) -> DominanceHierarchy:
        dominance_hierarchy_texts = self.rollout_dataframe['population.dominance_hierarchy'][::-1]
        for dominance_hierarchy_text in dominance_hierarchy_texts:
            dominance_hierarchy = DominanceHierarchy(dominance_hierarchy_text)
            if dominance_hierarchy.is_complete:
                return dominance_hierarchy
        return DominanceHierarchy(dominance_hierarchy_texts.iloc[0])

    @property
    def rollout_dataframe(self) -> pd.DataFrame:
        return misc.jsonl_to_dataframe(self.rollout_jsonl_path, dot_notation=True)


    @property
    def rankized_rollout_dataframe(self) -> pd.DataFrame:
        if not self.dominance_hierarchy.is_transitive:
            raise AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy
        if not self.dominance_hierarchy.is_complete:
            raise AttemptedCompleteOperationOnIncompleteDominanceHierarchy
        i_agent_by_rank = self.dominance_hierarchy.i_agent_by_rank
        rank_by_i_agent = tuple(i_agent_by_rank.index(i_agent)
                                for i_agent in self.dominance_hierarchy.i_agents)
        assert set(rank_by_i_agent) == set(i_agent_by_rank)

        def rename_agent_column(agent_column: str) -> str:
            i_agent_str, rest = re.fullmatch('^agent\.([0-9]+)\.(.+)$', agent_column).groups()
            rank = rank_by_i_agent[int(i_agent_str)]
            return f'rank.{rank}.{rest}'

        df = self.rollout_dataframe

        df.rename(columns={column: rename_agent_column(column)
                           for column in df.columns if column.startswith('agent.')},
                  inplace=True)

        return df

    @property
    def policy_path_by_name(self) -> dict[str, pathlib.Path]:
        return {
            policy_path.name: policy_path for policy_path in
            sorted(sorted((self.folder / 'policy_snapshots').iterdir())[-1].iterdir())
        }


    @staticmethod
    def get(mini_trek_descendant_path: str | os.PathLike) -> Trek:
        mini_trek_descendant_path = pathlib.Path(mini_trek_descendant_path)
        candidate_paths = itertools.chain((mini_trek_descendant_path,),
                                          mini_trek_descendant_path.parents)
        for candidate_path in candidate_paths:
            mini_trek_candidate = MiniTrek(candidate_path)
            if mini_trek_candidate.rollout_jsonl_path.exists():
                return mini_trek_candidate
        raise Exception

